﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using NRG.Models;

namespace NRG.Import
{
    public class BinReader
    {
        #region Properties

        private BinaryReader reader;
        public List<Line> Lines { get; private set; }
        public List<CodedPoint> Points { get; private set; }
        public List<Polygon> SelectionZones { get; private set; }
        public List<Polygon> DeselectionZones { get; private set; }
        public List<CloudPoint> FilteredPoints { get; private set; }
        public string PointCloudFile { get; private set; }

        #endregion

        #region Setup

        public BinReader(string filename)
        {
            reader = new BinaryReader(File.Open(filename, FileMode.Open, FileAccess.Read, FileShare.Read));
            Lines = new List<Line>();
            Points = new List<CodedPoint>();
            SelectionZones = new List<Polygon>();
            DeselectionZones = new List<Polygon>();
            FilteredPoints = new List<CloudPoint>();
            PointCloudFile = "";
        }

        #endregion

        #region Methods

        public bool ReadFile()
        {
            try
            {
                //Read the first string to ensure its a nrg file
                var checkString = reader.ReadString();
                if (checkString != "NRG User Data File Version 1.0")
                    return false;

                PointCloudFile = reader.ReadString();

                var finalPosition = reader.BaseStream.Length;

                while (reader.BaseStream.Position < finalPosition)
                {
                    var dataType = reader.ReadInt32();
                    switch ((DataTypes)dataType)
                    {
                        case DataTypes.Line:
                            ReadLines();
                            break;
                        case DataTypes.CodedPoint:
                            ReadPoints();
                            break;
                        case DataTypes.SelectionZone:
                            ReadSelectionZones();
                            break;
                        case DataTypes.DeselectionZone:
                            ReadDeselectionZones();
                            break;
                        case DataTypes.FilteredPoints:
                            ReadFilteredPoints();
                            break;
                    }
                }

                reader.Close();
                reader = null;

                return true;
            }
            catch
            {
                if (reader != null)
                    reader.Close();
                reader = null;

                return false;
            }
        }

        private void ReadLines()
        {
            var count = reader.ReadInt32();
            Lines = new List<Line>(count);

            for (int i = 0; i < count; i++)
            {
                var line = new Line();
                line.Code = reader.ReadString();
                line.IsPoly = reader.ReadBoolean();
                line.Color = System.Drawing.Color.FromArgb(reader.ReadByte(), reader.ReadByte(), reader.ReadByte());
                line.Thickness = reader.ReadDouble();

                var pointCount = reader.ReadInt32();
                line.Points = new List<Point3D>(pointCount);

                for (int j = 0; j < pointCount; j++)
                {
                    var point = new Point3D();
                    point.X = reader.ReadDouble();
                    point.Y = reader.ReadDouble();
                    point.Z = reader.ReadDouble();
                    line.Points.Add(point);
                }

                Lines.Add(line);
            }
        }

        private void ReadPoints()
        {
            var count = reader.ReadInt32();
            Points = new List<CodedPoint>(count);

            for (int i = 0; i < count; i++)
            {
                var point = new CodedPoint();
                point.Code = reader.ReadString();
                point.Color = System.Drawing.Color.FromArgb(reader.ReadByte(), reader.ReadByte(), reader.ReadByte());
                point.Thickness = reader.ReadDouble();
                point.X = reader.ReadDouble();
                point.Y = reader.ReadDouble();
                point.Z = reader.ReadDouble();
                Points.Add(point);
            }
        }

        private void ReadSelectionZones()
        {
            var count = reader.ReadInt32();
            SelectionZones = new List<Polygon>(count);

            for (int i = 0; i < count; i++)
            {
                var zone = new Polygon();
                zone.IsDeselectionZone = reader.ReadBoolean();
                zone.Filtered = reader.ReadBoolean();

                var pointCount = reader.ReadInt32();
                zone.Points = new List<Point3D>(pointCount);

                for (int j = 0; j < pointCount; j++)
                {
                    var point = new Point3D();
                    point.X = reader.ReadDouble();
                    point.Y = reader.ReadDouble();
                    point.Z = reader.ReadDouble();
                    zone.Points.Add(point);
                }
                SelectionZones.Add(zone);
            }
        }

        private void ReadDeselectionZones()
        {
            var count = reader.ReadInt32();
            DeselectionZones = new List<Polygon>(count);

            for (int i = 0; i < count; i++)
            {
                var zone = new Polygon();
                zone.IsDeselectionZone = reader.ReadBoolean();
                zone.Filtered = reader.ReadBoolean();

                var pointCount = reader.ReadInt32();
                zone.Points = new List<Point3D>(pointCount);

                for (int j = 0; j < pointCount; j++)
                {
                    var point = new Point3D();
                    point.X = reader.ReadDouble();
                    point.Y = reader.ReadDouble();
                    point.Z = reader.ReadDouble();
                    zone.Points.Add(point);
                }
                DeselectionZones.Add(zone);
            }
        }

        private void ReadFilteredPoints()
        {
            var count = reader.ReadInt32();
            FilteredPoints = new List<CloudPoint>(count);

            for (int i = 0; i < count; i++)
            {
                var point = new CloudPoint();
                point.X = reader.ReadDouble();
                point.Y = reader.ReadDouble();
                point.Z = reader.ReadDouble();
                point.R = reader.ReadByte();
                point.G = reader.ReadByte();
                point.B = reader.ReadByte();
                point.Intensity = reader.ReadUInt16();
                FilteredPoints.Add(point);
            }
        }

        #endregion
    }
}
