﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml;
using System.Xml.Schema;
using System.Xml.Linq;
using System.IO;
using NRG.Models;
using System.Windows.Forms;

namespace NRG.Import


{

    /*
     * Timble Tunnel Schema for TXL file format
     * https://www.trimble.com/schema/TunnelXML/6_0/TunnelXMLSchema-6.0.xsd
     *
     *   ------------------------------------------------------------------------------
     *   Some important notes from the Trimble schema if you can't be arsed to read it:
     *   ------------------------------------------------------------------------------
     *   All TunnelXML file angle values are in decimal degrees
     *   All TunnelXML file distance/measurement values are in metres regardless of NRG project settings
     *   All TunnelXML file grades values (gradients) are percentages
    */
    public class TrimbleTXLAlignment
    {

        public Alignment TXLAlignment = new Alignment();

        #region file section constants
        private const string HORIZONTAL = "HorizontalAlignment";
        private const string HORIZONTALIP = "HorizontalAlignmentPIs";

        private const string VERTICAL = "VerticalAlignment";
        private const string SHAPEPOSITION = "TemplatePositioning";
        private const string SHAPE = "TemplateRecord";
        #endregion

        #region horizontal ip constants
        private const string PICOORD = "PICoordinate";
        private const string START = "StartPI";
        private const string ELE = "PI";
        private const string SPIRALLENIN = "SpiralInLength";
        private const string SPIRALLENOUT = "SpiralInLength";
        private const string ARCLEN = "ArcLength";
        private const string ARCRAD = "ArcRadius";
        #endregion horizontal ip constants

        #region horizontal element constants
        private const string STRAIGHT = "Straight";
        private const string ARC = "Arc";
        private const string SPIRALIN = "EntrySpiral";
        private const string SPIRALOUT = "ExitSpiral";
        #endregion

        #region horizontal parameter constants
        private const string STARTCOORD = "StartCoordinate";
        private const string ENDCOORD = "EndCoordinate";


        private const string EAST = "East"; 
        private const string NORTH = "North";

        private const string STARTCH = "StartStation";
        private const string ENDCH = "EndStation";

        private const string  STARTBRG = "StartAzimuth";
        private const string ENDBRG = "EndAzimuth";

        private const string LENGTH = "Length";

        private const string STARTRAD = "StartRadius";
        private const string ENDRAD = "EndRadius";
        

        private const string DIR = "Direction";

        #endregion

        #region vertical element constants

        private const string STARTPT = "StartPoint";
        private const string ENDPT = "EndPoint";
        private const string VERTPOINT = "VerticalPoint";
        private const string VERTARC = "VerticalArc";
        private const string VERTPARA = "VerticalParabola";
        private const string VIP = "IntersectionPoint";
        private const string LEVELCH = "Station";
        private const string LEVELELEV = "Elevation";
        private const string GRADEIN = "GradeIn";
        private const string GRADEOUT = "GradeOut";


        #endregion vertical element constants

        #region general parameter constants
        private const string RAD = "Radius";
        private const string CH = "Station";
        #endregion general parameter contants

        public bool ReadFile(string filename = "", string InitDirectory = "")
        {

            if (File.Exists(filename) == false)
            {
                OpenFileDialog ofd = new OpenFileDialog();

                if (Directory.Exists(InitDirectory))
                {
                    ofd.InitialDirectory = InitDirectory;
                }

                ofd.Filter = "Trimble Tunnel Alignments| *.TXL; *.TXL;)";
                ofd.Title = "Import Trimble Tunnel Alignmemt";
                ofd.ShowDialog();
                filename = ofd.FileName;
            }

            if (File.Exists(filename) == false)
            {
                return false;
            }

            XDocument txl = XDocument.Load(filename);
            //txl = xdoc(filename);

            if (txl == null)
            {
                return false;
            }

            if (ParseFile(txl) == false)
            {
                return false;
            }

            return true;
        }

        private bool ParseFile(XDocument txl)
        {
          
            if (txl == null) return false;

            #region horizontal ips

            if (txl.Descendants(HORIZONTALIP).Any())
            {
                List<Maths.Geometry.HzIpTransition> ips = new List<Maths.Geometry.HzIpTransition>();

                foreach (XElement item in txl.Descendants(HORIZONTALIP))
                {

                    if (item.HasElements)
                    {
                        double LastEast = 0;
                        double lastNorth = 0;
                        double startCh = 0;

                        foreach (XElement he in item.Elements())
                        {
                            
                            Maths.Geometry.HzIpTransition ip = new Maths.Geometry.HzIpTransition();
                            switch (he.Name.ToString())
                            {
                                case START:
                                    ip = HorizontalIPSTart(he);
                                    startCh = ip.InitialChainage;
                                    LastEast = ip.IP1.X;
                                    lastNorth = ip.IP1.Y;
                                    break;
                                case ELE:
                                    
                                    ip = HorizontalIP(he);
                                    ip.IP1.X = LastEast;
                                    ip.IP1.Y = lastNorth;
                                    LastEast = ip.IP2.X;
                                    lastNorth = ip.IP2.Y;
                                    Maths.Geometry.HzIpTransition lastip = new Maths.Geometry.HzIpTransition();
                                    if (ips.Count > 0)
                                    {
                                        lastip = ips.Last();
                                        lastip.IP3.X = LastEast;
                                        lastip.IP3.Y = lastNorth;
                                    }

                                    if (ip.ArcRadius != 0)
                                    {
                                        ips.Add(ip);
                                    }
                                    
                                    break;
                            }


                        }

                    }


                }
                #region convert to nrg hz alignment  
                //do NRG string conversion here...
                int i = 0;
                double previousCh = 0;
                if (ips.Count > 0)
                
                {
                    HorizontalElement lastele = new HorizontalElement();
                    double lastch = 0;
                    Maths.Geometry.HzIpTransition lastip = new Maths.Geometry.HzIpTransition();
                    List<Maths.Geometry.HzIpResult> results = new List<Maths.Geometry.HzIpResult>();
                    foreach (Maths.Geometry.HzIpTransition ip in ips)
                        
                    {
                        HorizontalElement ele = new HorizontalElement();
                        i++;
                        Maths.Geometry.HzIpResult res = Maths.Geometry.SolveTransition(ip, false, true, true, true);


                        if (i == 1)
                        {
                            ele.Chainage = ip.InitialChainage;
                            ele.ChainageOrigin = ele.Chainage;
                            lastch = ele.Chainage;
                            ele.East = ip.IP1.X;
                            ele.North = ip.IP1.Y;
                            ele.Radius = 0;
                            ele.RadiusOrigin = 0;
                            ele.BrgOrigin = res.TangentSprialBearing;
                            ele.EastOrigin = ele.East;
                            ele.NorthOrigin = ele.North;
                            TXLAlignment.Horizontal.AddElement(ele);
                        }
                        else
                        {
                            
                            lastch = Maths.Vector.FnDistance(lastele.East, lastele.North, lastip.IP2.X,lastip.IP2.Y);
                        }
                        ele = new HorizontalElement();
                        ele.Chainage = previousCh + res.TangentSpiralChaiange - lastch;
                        ele.ChainageOrigin = previousCh + res.TangentSpiralChaiange - lastch;
                        ele.East = res.TangentSpralCoords.X;
                        ele.EastOrigin = ele.East;
                        ele.North = res.TangentSpralCoords.Y;
                        ele.NorthOrigin = ele.North;
                        ele.BrgOrigin = res.TangentSprialBearing;
                        ele.Radius = ip.ArcRadius;
                        ele.RadiusOrigin = 0;
                        ele.RL = ele.Radius * ip.SpiralLengthIn;
                        lastch = res.TangentSpiralChaiange;
                        previousCh = ele.Chainage;
                        TXLAlignment.Horizontal.AddElement(ele);

                        ele = new HorizontalElement();
                        ele.Chainage = previousCh + res.SpiralArcChaiange - lastch;
                        ele.ChainageOrigin = ele.Chainage;
                        ele.East = res.SpiralArcCoords.X;
                        ele.EastOrigin = ele.East;
                        ele.North = res.SpiralArcCoords.Y;
                        ele.NorthOrigin = ele.North;
                        ele.BrgOrigin = res.SpralArcBearing;
                        ele.Radius = ip.ArcRadius;
                        ele.RadiusOrigin = ele.Radius;
                        ele.RL = 0;
                        previousCh = ele.Chainage;
                        lastch = res.SpiralArcChaiange;

                        TXLAlignment.Horizontal.AddElement(ele);

                        ele = new HorizontalElement();
                        ele.Chainage = previousCh + res.ArcSpiralChaiange - lastch;
                        ele.ChainageOrigin = previousCh+ res.SpiralTangentChaiange - lastch;
                        ele.East = res.SpiralTangentCoords.X;
                        ele.EastOrigin = ele.East;
                        ele.North = res.SpiralTangentCoords.Y;
                        ele.NorthOrigin = ele.North;
                        ele.BrgOrigin = res.SpiralTangentBearing;
                        ele.Radius = ip.ArcRadius;
                        ele.RadiusOrigin = ele.Radius;
                        ele.RL = ele.Radius * ip.SpiralLengthOut;
                        lastch = res.ArcSpiralChaiange;
                        previousCh = ele.Chainage;
                        TXLAlignment.Horizontal.AddElement(ele);

                        //


                        ele = new HorizontalElement();
                        ele.Chainage = previousCh + res.SpiralTangentChaiange - lastch;
                        ele.ChainageOrigin = previousCh +res.SpiralTangentChaiange - lastch;
                        ele.East = res.SpiralTangentCoords.X;
                        ele.EastOrigin = ele.East;
                        ele.North = res.SpiralTangentCoords.Y;
                        ele.NorthOrigin = ele.North;
                        ele.BrgOrigin = res.SpiralTangentBearing;
                        ele.Radius = 0;
                        ele.RadiusOrigin = 0;
                        ele.RL = 0;
                        lastch = ele.Chainage;
                        TXLAlignment.Horizontal.AddElement(ele);

                        lastch = res.SpiralTangentChaiange;

                        previousCh = ele.Chainage;

                        if (ip == ips.Last())
                        {
                            HorizontalElement ele2 = new HorizontalElement();
                            ele2.East = ip.IP3.X;
                            ele2.North = ip.IP3.Y;
                            ele2.Radius = 0;
                            ele2.RadiusOrigin = 0;
                            ele2.RL = 0;
                            ele2.BrgOrigin = ele.BrgOrigin;
                            ele2.Chainage = Maths.Vector.FnDistance(ele.East, ele.North, ele2.East, ele2.North) + ele.Chainage;
                            ele2.ChainageOrigin = ele2.Chainage;
                            TXLAlignment.Horizontal.AddElement(ele2);

                        }
                        lastip = new Maths.Geometry.HzIpTransition();
                        lastip = ip;
                        lastele = new HorizontalElement();
                        lastele = ele;
                        if (res != null)
                        {
                            results.Add(res);
                        }
                        else
                        {
                            Console.WriteLine("Not OK");

                        }


                    }
                    
                }
                #endregion convert to nrg hz alignment 


            }


            #endregion horizontal ips

            #region horizontal alignment (tangent points)

            //I'm pretty happy with all this. There's some stuff which I think maybe related to 
            //back to back or spiral to spiral geometry that I haven't catered for, but we might 
            //just have to deal with that as problems arise.
            foreach (XElement item in txl.Descendants(HORIZONTAL))
            {
                if (item.HasElements)
                {
                    HorizontalElement ele = new HorizontalElement();

                    foreach (XElement he in item.Elements())
                    {
                        switch (he.Name.ToString())
                        {
                            case STRAIGHT:
                                ele = HzStraight(he);
                                break;
                            case ARC:
                                ele = HzArc(he);
                                break;
                            case SPIRALIN:
                                ele = HzEntrySpiral(he);
                                break;
                            case SPIRALOUT:
                                ele = HzExitSpiral(he);
                                break;
                            default:
                                ele = null;
                                break;
                        }
                        if (ele !=null)
                        { TXLAlignment.Horizontal.AddElement(ele); }
                        if (he == item.LastNode)
                        {
                            HorizontalElement hz2 = new HorizontalElement();
                            hz2 = CreateLastElement(he);
                            TXLAlignment.Horizontal.AddElement(hz2);
                        }
                        
                    }
                
                }

            }

            #endregion horizontal alignment

            #region vertical alignment
            //this is where stuff gets interesting. It seems we can mix and match vertical intersection
            //points and vertical tangent points. I'd love to meet the committee that came up with this 
            //horse with two humps

            foreach (XElement item in txl.Descendants(VERTICAL))
            {
                if (item.HasElements)
                {
                    
                    foreach (XElement ve in item.Elements())
                    {
                        VerticalElement ele = new VerticalElement();
                        switch (ve.Name.ToString())
                        {
                            case VERTPOINT:
                                ele = VertPoint(ve);
                                break;
                            case VERTARC:
                                VertArc(ve, TXLAlignment.Vertical);
                                ele = null;
                                break;
                            //case VERTPARA:
                            //    break;
                            default:
                                ele = null;
                                break;
                        }
                        if (ele != null)
                        {
                            TXLAlignment.Vertical.AddElement(ele);
                        }
                    }

                }
            }

            #endregion vertical alignment


            return TXLAlignment.Horizontal.Elements.Count > 0; 
        }


        //re-factoring

        #region extracting horizontal ip

        private Maths.Geometry.HzIpTransition HorizontalIPSTart(XElement txl)
        {
            Maths.Geometry.HzIpTransition ele = new Maths.Geometry.HzIpTransition();

            ele.InitialChainage = TXLValue(txl, STARTCH);
            ele.IP1.X = TXLValue(txl, PICOORD, EAST);
            ele.IP1.Y = TXLValue(txl, PICOORD, NORTH);
            return ele;
        }

        private Maths.Geometry.HzIpTransition HorizontalIP(XElement txl)
        {
            Maths.Geometry.HzIpTransition ele = new Maths.Geometry.HzIpTransition();

            ele.IP2.X = TXLValue(txl, PICOORD, EAST);
            ele.IP2.Y = TXLValue(txl, PICOORD, NORTH);
            ele.ArcLength = TXLValue(txl, ARCLEN);
            ele.ArcRadius = TXLValue(txl, ARCRAD);
            ele.SpiralLengthIn = TXLValue(txl, SPIRALLENIN);
            ele.SpiralLengthOut = TXLValue(txl, SPIRALLENOUT);
            return ele;
            
        }

        private HorizontalElement HzStraight(XElement txl)
        {
            HorizontalElement ele = new HorizontalElement();

            ele.Chainage = TXLValue(txl, STARTCH);
            ele.ChainageOrigin = ele.Chainage;

            ele.East = TXLValue(txl, STARTCOORD, EAST);
            ele.North = TXLValue(txl, STARTCOORD, NORTH);
            ele.EastOrigin = ele.East;
            ele.NorthOrigin = ele.North;

            ele.BrgOrigin = TXLValue(txl, STARTBRG);

            ele.Radius = TXLValue(txl, RAD); 

            ele.Length = TXLValue(txl, LENGTH);

            ele.RL = 0;

            return ele;
        }

        private HorizontalElement HzArc(XElement txl)
        {
            HorizontalElement ele = new HorizontalElement();

            ele.Chainage = TXLValue(txl, STARTCH);
            ele.ChainageOrigin = ele.Chainage;

            ele.East = TXLValue(txl, STARTCOORD, EAST);
            ele.North = TXLValue(txl, STARTCOORD, NORTH);
            ele.EastOrigin = ele.East;
            ele.NorthOrigin = ele.North;

            ele.BrgOrigin = TXLValue(txl, STARTBRG);

            ele.Radius = TXLValue(txl, RAD);
            ele.RadiusOrigin = ele.Radius;

            ele.Length = TXLValue(txl, LENGTH);

            ele.RL = 0;

            return ele;
        }

        private HorizontalElement HzEntrySpiral(XElement txl)
        {
            HorizontalElement ele = new HorizontalElement();

            ele.ElementType = HorizontalElementType.Spiral;
            ele.Chainage = TXLValue(txl, STARTCH);
            ele.ChainageOrigin = ele.Chainage;

            ele.East = TXLValue(txl, STARTCOORD, EAST);
            ele.North = TXLValue(txl, STARTCOORD, NORTH);
            ele.EastOrigin = ele.East;
            ele.NorthOrigin = ele.North;

            ele.BrgOrigin = TXLValue(txl, STARTBRG);

            ele.Radius = TXLValue(txl, ENDRAD);
            ele.RadiusOrigin = 0.0f;

            ele.Length = TXLValue(txl, LENGTH);

            ele.RL = ele.Length * ele.Radius; 

            return ele;
        }

        private HorizontalElement HzExitSpiral(XElement txl)
        {
            HorizontalElement ele = new HorizontalElement();

            ele.ElementType = HorizontalElementType.Spiral;
            ele.Chainage = TXLValue(txl, STARTCH);
            ele.ChainageOrigin = TXLValue(txl, ENDCH);

            ele.East = TXLValue(txl, STARTCOORD, EAST);
            ele.North = TXLValue(txl, STARTCOORD, NORTH);
            ele.EastOrigin = TXLValue(txl, ENDCOORD, EAST);
            ele.NorthOrigin = TXLValue(txl, ENDCOORD, NORTH);

            ele.BrgOrigin = TXLValue(txl, ENDBRG);

            ele.Radius = TXLValue(txl, STARTRAD);
            ele.RadiusOrigin = ele.Radius;

            ele.Length = TXLValue(txl, LENGTH);

            ele.RL = ele.Length * ele.Radius;

            return ele;
        }

        private HorizontalElement CreateLastElement(XElement txl)
        {
            HorizontalElement ele = new HorizontalElement();

            ele.East = TXLValue(txl,ENDCOORD, EAST);
            ele.North = TXLValue(txl, ENDCOORD, NORTH);
            ele.EastOrigin = TXLValue(txl, ENDCOORD, EAST);
            ele.NorthOrigin = TXLValue(txl, ENDCOORD, NORTH);
            ele.BrgOrigin= TXLValue(txl, ENDBRG);
            ele.Chainage = TXLValue(txl, ENDCH);
            ele.ChainageOrigin = TXLValue(txl, ENDCH);
            ele.Radius = 0;
            ele.RL = 0;

            return ele;
        }
        #endregion
    
        #region extracting vertical
        VerticalElement VertPoint(XElement txl)
        {
            VerticalElement ele = new VerticalElement();

            ele.ElementType = VerticalElementType.Straight;
            ele.Chainage = TXLValue(txl,VIP,LEVELCH);
            ele.Level = TXLValue(txl, VIP, LEVELELEV);
            ele.Radius = 0;

            return ele;
        }

        void VertArc(XElement txl , VerticalAlignment va)
        {
            VerticalElement ele = new VerticalElement();
            ele.ElementType = VerticalElementType.Parabola;
            ele.Radius = TXLValue(txl, RAD);

            ele.Chainage = TXLValue(txl,STARTPT,LEVELCH);
            ele.Level = TXLValue(txl, STARTPT, LEVELELEV);

            va.AddElement(ele);

            ele = new VerticalElement();
            ele.Chainage = TXLValue(txl, ENDPT, LEVELCH);
            ele.Level = TXLValue(txl, ENDPT, LEVELELEV);
            ele.Radius = 0;

            va.AddElement (ele);

            
        }
        #endregion extracting vertical

        #region parsing methods
        private double TXLValue(XElement txl, string key)
        {
            double val = 0;
            if (txl.Descendants(key).Any())
            {
                double.TryParse(txl.Element(key).Value.ToString(), out val);
                if (string.Equals(key, STARTBRG) || string.Equals(key, ENDBRG))
                {
                    val = val / 180 * Math.PI;
                }
                else if (string.Equals(key, RAD) || string.Equals(key, STARTRAD) || string.Equals(key, ENDRAD))

                {
                    val = Math.Abs(val);
                    if (txl.Descendants(DIR).Any())
                    {
                        if (string.Equals(txl.Element(DIR), "Left"))
                        {
                            val = -val;
                        }
                    }
                }
            }
            return val;
        }

        private double TXLValue(XElement txl, string key1, string key2)
        {
            double val = 0;
            if (txl.Descendants(key1).Any())
            {
                if (txl.Descendants(key1).Descendants(key2).Any())
                {
                    double.TryParse(txl.Element(key1).Element(key2).Value.ToString(), out val);
                    if (string.Equals(key2, STARTBRG) || string.Equals(key2, ENDBRG))
                    {
                        val = val / 180 * Math.PI;
                    }
                    else if (string.Equals(key2, RAD) || string.Equals(key2, STARTRAD) || string.Equals(key2, ENDRAD))

                    {
                        val = Math.Abs(val);
                        if (txl.Descendants(DIR).Any())
                        {
                            if (string.Equals(txl.Element(DIR), "Left"))
                            {
                                val = -val;
                            }
                        }
                    }
                }
            }
            return val;
        }
        #endregion


    }

}




