﻿using System;
using System.Drawing.Design;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Globalization;
using NRG.Services;
using NRG.Maths;
using TriangleNet.Geometry;

namespace NRG.Models

{

    


    [TypeConverter(typeof(LineConverter))]
    public class Line
    {
        #region Properties

        private int id;

        [Category("Line"), Description("The index ID for the line")]
        public int ID
        {
            get { return id; }
        }

        [Category("Line"), Description("The line code used for the line")]
        public string Code { get; set; }

        public string startLabel = "";

        public string StartLineCode
        {
            get
            {
                if (startLabel == "")
                    return Code;
                else
                    return startLabel + " " + Code;
            }
        }

        [Category("Line"), Description("Whether the line will close")]
        public bool IsPoly { get; set; }

        [Category("Coordinates"), Description("A list of coordinate points for the line")]
        public List<Point3D> Points { get; set; }

        [Category("Coordinates"), Description("The number of coordinate points for the line")]
        public int PointCount
        {
            get { return Points.Count; }
        }

        public System.Drawing.Color Color = System.Drawing.Color.Black;

        [Category("Color"), Description("The color of the line")]
        public System.Drawing.Color LineColor
        {
            get { return Color; }
        }

        [Category("Line"), Description("The thickness of the line")]
        public double Thickness { get; set; }

        #endregion

        #region Setup

        public Line()
        {
            this.id = 0;
            this.IsPoly = false;
            this.Code = "";
            this.Points = new List<Point3D>();
            this.Thickness = 0.2;
        }

        public Line(string code, int id, bool isPoly = false, int r = 0, int g = 0, int b = 0)
        {
            this.Code = code;
            this.id = id;
            this.IsPoly = isPoly;
            this.Points = new List<Point3D>();
            this.Color = System.Drawing.Color.FromArgb(r, g, b);
            this.Thickness = 0.2;
        }

        public Line(int id)
        {
            this.id = id;
            this.IsPoly = false;
            this.Code = "";
            this.Points = new List<Point3D>();
            this.Thickness = 0.2;
        }

        #endregion

        #region Methods

        public Line Copy()
        {
            var newLine = new Line(this.ID)
            {
                Code = this.Code,
                IsPoly = this.IsPoly,
                Color = this.Color,
                startLabel = this.startLabel,
                Thickness = this.Thickness,
                id = this.id,
                Points = new List<Point3D>()
            };

            foreach (var point in this.Points)
            {
                newLine.Points.Add(point.Copy());
            }
            return newLine;
        }

        public void SetID(int newId)
        {
            this.id = newId;
        }

        #endregion
    }

    [TypeConverter(typeof(PolygonConverter))]
    public class Polygon
    {
        #region Properties

        private int id;

        [Category("Polygon"), Description("The index ID for the polygon")]
        public int ID
        {
            get { return id; }
        }

        [Category("Polygon"), Description("Whether the polygon is a deselection zone or not")]
        public bool IsDeselectionZone { get; set; }

        [Category("Coordinates"), Description("A list of coordinate points for the polygon")]
        public List<Point3D> Points { get; set; }

        [Category("Coordinates"), Description("The number of coordinate points for the polygon")]
        public int PointCount
        {
            get { return Points.Count; }
        }

        public bool RemovedDuplicated = false;
        public bool Filtered = false;
        public Bounds SquareBounds = new Bounds();

        #endregion

        #region Setup

        public Polygon()
        {
            this.id = 0;
            this.IsDeselectionZone = false;
            this.Points = new List<Point3D>();
        }

        public Polygon(int id, bool isDeslectionZone = false)
        {
            this.id = id;
            this.IsDeselectionZone = isDeslectionZone;
            this.Points = new List<Point3D>();
        }

        public Polygon(int id)
        {
            this.id = id;
        }

        #endregion

        #region Methods

        public Polygon Copy()
        {
            var newPoly = new Polygon(this.ID)
            {
                IsDeselectionZone = this.IsDeselectionZone,
                RemovedDuplicated = this.RemovedDuplicated,
                SquareBounds = this.SquareBounds.Copy(),
                Points = new List<Point3D>(),
                Filtered = this.Filtered
            };

            foreach (var point in this.Points)
            {
                newPoly.Points.Add(point.Copy());
            }

            return newPoly;
        }

        public bool InPoly2D(double x, double y, double z)
        {
            bool isInPoly = false;
            int il = 0;
            int i = 0;
            //start and enpoints of a line seg;
            double sx = 0;
            double sy = 0;
            double ex = 0;
            double ey = 0;
            int crossings = 0;

            int numLasso = Points.Count;

            //if we have less than two points we don't have a polygon
            if (numLasso < 2)
                return false;

            crossings = 0;
            //go through each line segment in the polygon
            for (i = 0; i < numLasso; i++)
            {
                //get the start and end points of the line relative to origin x,y
                il = i - 1;
                if (il == -1)
                    il = numLasso - 1;
                sx = Points[il].X - x;
                sy = Points[il].Y - y;
                ex = Points[i].X - x;
                ey = Points[i].Y - y;
                if (sx == 0 && sy == 0)
                    return true;

                if (ex == 0 && ey == 0)
                    return true;

                //check if the line segment sx, sy - ex, ey straddles the x axis
                if (ey > 0 && sy <= 0)
                {
                    //calculate the intersetion with x axis
                    double ix = (ex * sy - sx * ey) / (double)(sy - ey);
                    //if this intersection is strictly +ve then we have a crossing
                    if (ix > 0)
                        crossings += 1;
                }

                if (sy > 0 && ey <= 0)
                {
                    //calculate the intersection with the x axis
                    double ix = (ex * sy - sx * ey) / (double)(sy - ey);
                    //if this intersection is strictly +ve then we have a crossing
                    if (ix > 0)
                        crossings += 1;
                }
            }

            //if there is an odd number of crossings then the point is in the poly
            if ((crossings % 2) == 1)
                isInPoly = true;
            else
                isInPoly = false;

            return isInPoly;
        }

        public Bounds Get2DSquareBounds(Bounds originalBounds)
        {
            var squareBounds = new Bounds();

            //Search through all points and find the max x, min x, max y, min y and set z to be 0, 0
            if (Points.Count <= 2)
                return squareBounds;

            double minX = Points[0].X, maxX = Points[0].X, minY = Points[0].Y, maxY = Points[0].Y;

            for (int i = 0; i < Points.Count;)
            {
                var point = Points[i];
                if (i >= 1)
                {
                    var lastPoint = Points[i - 1];
                    if (lastPoint.X == point.X && lastPoint.Y == point.Y && lastPoint.Z == point.Z)
                    {
                        //Remove duplicate
                        Points.RemoveAt(i);
                        continue;
                    }
                }

                if (point.X < minX)
                    minX = point.X;
                else if (point.X > maxX)
                    maxX = point.X;

                if (point.Y < minY)
                    minY = point.Y;
                else if (point.Y > maxY)
                    maxY = point.Y;

                i++;
            }
            RemovedDuplicated = true;
            squareBounds = new Bounds(minX, minY, originalBounds.Min.Z, maxX, maxY, originalBounds.Max.Z);
            SquareBounds = squareBounds;
            return squareBounds;
        }

        public Bounds Get2DSquareBounds(double minZ, double maxZ)
        {
            var squareBounds = new Bounds();

            //Search through all points and find the max x, min X, max Y
            if (Points.Count <= 2)
                return squareBounds;

            double minX = Points[0].X, maxX = Points[0].X, minY = Points[0].Y, maxY = Points[0].Y;

            for (int i = 0; i < Points.Count;)
            {
                var point = Points[i];
                if (i >= 1)
                {
                    var lastPoint = Points[i - 1];
                    if (lastPoint.X == point.X && lastPoint.Y == point.Y && lastPoint.Z == point.Z)
                    {
                        //Remove duplicate
                        Points.RemoveAt(i);
                        continue;
                    }
                }

                if (point.X < minX)
                    minX = point.X;
                else if (point.X > maxX)
                    maxX = point.X;

                if (point.Y < minY)
                    minY = point.Y;
                else if (point.Y > maxY)
                    maxY = point.Y;

                i++;
            }
            RemovedDuplicated = true;
            squareBounds = new Bounds(minX, minY, minZ, maxX, maxY, maxZ);
            SquareBounds = squareBounds;
            return squareBounds;
        }

        public void SetBoundLines(ref float[] vertices, ref float[] colors, double offsetX, double offsetY, double offsetZ, float r = 1.0f, float g = 0.953f, float b = 0.0588f)
        {
            if (Points.Count != 4 || SquareBounds == null)
                return;

            vertices = new float[72];
            colors = new float[72];
            int index = 0;

            var minZ = (float)(SquareBounds.Min.Z - offsetZ);
            var maxZ = (float)(SquareBounds.Max.Z - offsetZ);

            float[,] points = new float[,] {
                { (float)(Points[0].X - offsetX), (float)(Points[0].Y - offsetY) },
                { (float)(Points[1].X - offsetX), (float)(Points[1].Y - offsetY) },
                { (float)(Points[2].X - offsetX), (float)(Points[2].Y - offsetY) },
                { (float)(Points[3].X - offsetX), (float)(Points[3].Y - offsetY) } };

            //1 - 2, 2 - 3, 3 - 4, 4 - 1 Top
            for (int i = 0; i < 4; i++)
            {
                if (i == 3)
                {
                    AddValue(ref vertices, ref colors, ref index, points[3, 0], points[3, 1], maxZ, r, g, b);
                    AddValue(ref vertices, ref colors, ref index, points[0, 0], points[0, 1], maxZ, r, g, b);
                }
                else
                {
                    AddValue(ref vertices, ref colors, ref index, points[i, 0], points[i, 1], maxZ, r, g, b);
                    AddValue(ref vertices, ref colors, ref index, points[i + 1, 0], points[i + 1, 1], maxZ, r, g, b);
                }
            }

            //1 - 2, 2 - 3, 3 - 4, 4 - 1 Bot
            for (int i = 0; i < 4; i++)
            {
                if (i == 3)
                {
                    AddValue(ref vertices, ref colors, ref index, points[3, 0], points[3, 1], minZ, r, g, b);
                    AddValue(ref vertices, ref colors, ref index, points[0, 0], points[0, 1], minZ, r, g, b);
                }
                else
                {
                    AddValue(ref vertices, ref colors, ref index, points[i, 0], points[i, 1], minZ, r, g, b);
                    AddValue(ref vertices, ref colors, ref index, points[i + 1, 0], points[i + 1, 1], minZ, r, g, b);
                }
            }

            //1 - 1, 2 - 2, 3 - 3, 4 - 4 Top - Bot
            for (int i = 0; i < 4; i++)
            {
                AddValue(ref vertices, ref colors, ref index, points[i, 0], points[i, 1], maxZ, r, g, b);
                AddValue(ref vertices, ref colors, ref index, points[i, 0], points[i, 1], minZ, r, g, b);
            }
        }

        private void AddValue(ref float[] vertices, ref float[] colors, ref int index, float x, float y, float z, float r, float g, float b)
        {
            vertices[index] = x;
            colors[index++] = r;

            vertices[index] = y;
            colors[index++] = g;

            vertices[index] = z;
            colors[index++] = b;
        }

        public void SetID(int newId)
        {
            this.id = newId;
        }

        #endregion
    }

    public class ClearancePolygon
    {
        #region Properties

        public bool IsOverBreak { get; set; }
        public List<Point2D> Points { get; set; }

        public Bounds SquareBounds
        {
            get
            {
                var bounds = new Bounds();

                if (Points == null)
                    return bounds;

                foreach (var point in Points)
                {
                    bounds.Update(point);
                }

                return bounds;
            }
        }

        #endregion

        #region Setup

        public ClearancePolygon(bool isOverBreak)
        {
            IsOverBreak = isOverBreak;
            Points = new List<Point2D>();
        }

        #endregion

        #region Methods

        public ClearancePolygon Copy()
        {
            var newPoly = new ClearancePolygon(IsOverBreak)
            {
                Points = new List<Point2D>()
            };

            foreach (var point in Points)
                newPoly.Points.Add(point.Copy());

            return newPoly;
        }

        public List<Point2D> CalcShadingIntersectLines(List<Point2D> shadingLines)
        {
            var intersectedLines = new List<Point2D>();

            //Check for intersections for each line between each line segment of the polygon

            //Where an intersection is found add it as the start line , and find the next intersection to add as the end line
            for (int i = 0; i < shadingLines.Count - 1; i += 2)
            {
                var lineStart = shadingLines[i];
                var lineEnd = shadingLines[i + 1];
                double ipX = 0, ipY = 0;

                //double brg = 0, dist = 0;
                //Trig.RPC(lineStart.X, lineStart.Y, lineEnd.X, lineEnd.Y, ref brg, ref dist);

                var intersections = new List<Point2D>();

                //Check for intersections to next point
                for (int j = 0; j < Points.Count - 1; j++)
                {
                    var segStart = Points[j];
                    var segEnd = Points[j + 1];

                    if (Intersection.IntersectLines(lineStart.X, lineStart.Y, lineEnd.X, lineEnd.Y, segStart.X, segStart.Y, segEnd.X, segEnd.Y, ref ipX, ref ipY))
                        intersections.Add(new Point2D(ipX, ipY));


                    //var segBrg = Trig.RPC(segStart.X, segStart.Y, segEnd.X, segEnd.Y);

                    //var iPoint = Intersection.IntersectTwoBrgLines(lineStart.X, lineStart.Y, brg, segStart.X, segStart.Y, segBrg);
                    //if (Intersection.FnInBox1(lineStart.X, lineStart.Y, lineEnd.X, lineEnd.Y, iPoint.X, iPoint.Y) && Intersection.FnInBox1(segStart.X, segStart.Y, segEnd.X, segEnd.Y, iPoint.X, iPoint.Y))
                    //{
                    //    intersections.Add(iPoint);
                    //}
                    //var text = "";
                    //foreach (var point in Points)
                    //{
                    //    var test = "\"\", " + point.X + ", " + point.Y + ", " + "-999, " + "\"\"";
                    //    text += test + "\r\n";
                    //}
                    //System.IO.File.WriteAllText("test.txt", text);
                }

                if (intersections.Count % 2 != 0)
                    continue;

                for (int j = 0; j < intersections.Count;)
                {
                    int closestIndex = 0;
                    double distanceFrom = double.MaxValue;

                    for (int k = 0; k < intersections.Count; k++)
                    {
                        var d = Vector.FnDistance(lineStart.X, lineStart.Y, intersections[k].X, intersections[k].Y);
                        if (d < distanceFrom)
                        {
                            distanceFrom = d;
                            closestIndex = k;
                        }
                    }

                    //Make sure center point falls inside the polygon
                    var cX = (intersections[0].X + intersections[closestIndex].X) / 2;
                    var cY = (intersections[0].Y + intersections[closestIndex].Y) / 2;

                    //if (InPoly2D(cX, cY, 0))
                    //{
                        //intersectedLines.Add(intersections[0]);
                        intersectedLines.Add(intersections[closestIndex]);
                    //}

                    //Make sure we remove this one first or else index will change
                    intersections.RemoveAt(closestIndex);
                    //intersections.RemoveAt(0);
                }
            }

            return intersectedLines;
        }

        public double CalcArea()
        {
            double area = 0;
            Point2D xy1, xy2;

            for (int i = 0; i < Points.Count - 1; i++)
            {
                xy1 = Points[i];
                xy2 = Points[i + 1];

                area += ((xy2.X - xy1.X) * (xy2.Y + xy1.Y));
            }

            //Add last part
            xy1 = Points.Last();
            xy2 = Points[0];

            area += ((xy2.X - xy1.X) * (xy2.Y + xy1.Y));

            return area;
        }

        public bool InPoly2D(double x, double y, double z)
        {
            bool isInPoly = false;
            int il = 0;
            int i = 0;
            //start and enpoints of a line seg;
            double sx = 0;
            double sy = 0;
            double ex = 0;
            double ey = 0;
            int crossings = 0;

            int numLasso = Points.Count;

            //if we have less than two points we don't have a polygon
            if (numLasso < 2)
                return false;

            crossings = 0;
            //go through each line segment in the polygon
            for (i = 0; i < numLasso; i++)
            {
                //get the start and end points of the line relative to origin x,y
                il = i - 1;
                if (il == -1)
                    il = numLasso - 1;
                sx = Points[il].X - x;
                sy = Points[il].Y - y;
                ex = Points[i].X - x;
                ey = Points[i].Y - y;
                if (sx == 0 && sy == 0)
                    return true;

                if (ex == 0 && ey == 0)
                    return true;

                //check if the line segment sx, sy - ex, ey straddles the x axis
                if (ey > 0 && sy <= 0)
                {
                    //calculate the intersetion with x axis
                    double ix = (ex * sy - sx * ey) / (double)(sy - ey);
                    //if this intersection is strictly +ve then we have a crossing
                    if (ix > 0)
                        crossings += 1;
                }

                if (sy > 0 && ey <= 0)
                {
                    //calculate the intersection with the x axis
                    double ix = (ex * sy - sx * ey) / (double)(sy - ey);
                    //if this intersection is strictly +ve then we have a crossing
                    if (ix > 0)
                        crossings += 1;
                }
            }

            //if there is an odd number of crossings then the point is in the poly
            if ((crossings % 2) == 1)
                isInPoly = true;
            else
                isInPoly = false;

            return isInPoly;
        }

        public List<Point2D> CalcSolidFillTriangles()
        {
            var trianglePoints = new List<Point2D>();

            if (Points.Count <= 2)
                return trianglePoints;

            if (Points.Count == 3)
            {
                trianglePoints.Add(Points[0]);
                trianglePoints.Add(Points[1]);
                trianglePoints.Add(Points[2]);
                return trianglePoints;
            }

            var poly = new TriangleNet.Geometry.Polygon(Points.Count);

            foreach (var point in Points)
                poly.Add(new TriangleNet.Geometry.Vertex(point.X, point.Y));

            var constraintOptions = new TriangleNet.Meshing.ConstraintOptions()
            {
                ConformingDelaunay = true,
                Convex = false,
            };

            var qualityOptions = new TriangleNet.Meshing.QualityOptions()
            {
                MaximumAngle = 360,
                MinimumAngle = 0,
            };



            var mesh = poly.Triangulate(constraintOptions, qualityOptions);
            foreach(var triangle in mesh.Triangles)
            {
                var point1 = Points[triangle.GetVertexID(0)];
                if (!InPoly2D(point1.X, point1.Y, 0))
                    continue;

                var point2 = Points[triangle.GetVertexID(1)];
                if (!InPoly2D(point2.X, point2.Y, 0))
                    continue;

                var point3 = Points[triangle.GetVertexID(2)];
                if (!InPoly2D(point3.X, point3.Y, 0))
                    continue;

                //If the center of the triange is not in the polygon its an invalid triangle so go to next
                var centroid = new Point2D((point1.X + point2.X + point3.X) / 3, (point1.Y + point2.Y + point3.Y) / 3);
                if (!InPoly2D(centroid.X, centroid.Y, 0))
                    continue;

                trianglePoints.Add(point1);
                trianglePoints.Add(point2);
                trianglePoints.Add(point3);
            }

            //var inputPoints = new TriangleNet.Geometry.InputGeometry();

            //foreach (var point in Points)
            //    inputPoints.AddPoint(point.X, point.Y);
            //var triangleMesh = new TriangleNet.Mesh(null);
            //triangleMesh.

            //triangleMesh.Behavior.Algorithm = TriangleNet.TriangulationAlgorithm.Dwyer;
            //triangleMesh.Behavior.ConformingDelaunay = true;
            //triangleMesh.Behavior.Convex = false;
            //triangleMesh.Behavior.MinAngle = 0;
            //triangleMesh.Behavior.MaxAngle = 360;
            //triangleMesh.Behavior.NoHoles = true;
            //triangleMesh.Behavior.NoBisect = 1;
            //triangleMesh.Behavior.Quality = true;

            //triangleMesh.Triangulate(inputPoints);
            //triangleMesh.Renumber(TriangleNet.NodeNumbering.Linear);

            //foreach (var triangle in triangleMesh.Triangles)
            //{
            //    var point1 = Points[triangle.P0];
            //    if (!InPoly2D(point1.X, point1.Y, 0))
            //        continue;

            //    var point2 = Points[triangle.P1];
            //    if (!InPoly2D(point2.X, point2.Y, 0))
            //        continue;

            //    var point3 = Points[triangle.P2];
            //    if (!InPoly2D(point3.X, point3.Y, 0))
            //        continue;

            //    //If center of triangle is not in the polygon its an invalid triangle so go to next
            //    var centroid = new Point2D((point1.X + point2.X + point3.X) / 3, (point1.Y + point2.Y + point3.Y) / 3);
            //    if (!InPoly2D(centroid.X, centroid.Y, 0))
            //        continue;

            //    trianglePoints.Add(point1);
            //    trianglePoints.Add(point2);
            //    trianglePoints.Add(point3);

            return trianglePoints;
        }

        #endregion
    }

    [TypeConverter(typeof(Point3DConverter))]
    [Serializable]
    public class Point3D : Point2D
    {
        #region Properties

        [Category("Point"), Description("The z value of the point")]
        public double Z { get; set; }

        #endregion

        #region Setup

        public Point3D()
        {

        }

        public Point3D(double x, double y, double z)
        {
            this.X = x;
            this.Y = y;
            this.Z = z;
        }

        #endregion

        #region Operators

        public static Point3D operator -(Point3D left, Point3D right)
        {
            return new Point3D(left.X - right.X, left.Y - right.Y, left.Z - right.Z);
        }

        public static Point3D operator -(Point3D left, double right)
        {
            return new Point3D(left.X - right, left.Y - right, left.Z - right);
        }

        public static Point3D operator +(Point3D left, Point3D right)
        {
            return new Point3D(left.X + right.X, left.Y + right.Y, left.Z + right.Z);
        }

        public static Point3D operator +(Point3D left, double right)
        {
            return new Point3D(left.X + right, left.Y + right, left.Z + right);
        }

        public static Point3D operator *(Point3D left, double right)
        {
            return new Point3D(left.X * right, left.Y * right, left.Z * right);
        }

        public static Point3D operator *(double left, Point3D right)
        {
            return new Point3D(left * right.X, left * right.Y, left * right.Z);
        }

        public static Point3D operator /(Point3D left, double right)
        {
            return new Point3D(left.X / right, left.Y / right, left.Z / right);
        }

        #endregion

        #region Methods

        public override double Length()
        {
            return Math.Sqrt(X * X + Y * Y + Z * Z);
        }

        public override double SquaredLength()
        {
            return X * X + Y * Y + Z * Z;
        }

        public double DistanceTo(Point3D point)
        {
            return (this - point).Length();
        }

        public double SquaredDistanceTo(Point3D point)
        {
            return (this - point).SquaredLength();
        }

        public override double MaxValue()
        {
            return Math.Max(X, Math.Max(Y, Z));
        }

        public override double MinValue()
        {
            return Math.Min(X, Math.Min(Y, Z));
        }

        public Point2D ToPoint2D()
        {
            return new Point2D()
            {
                X = this.X,
                Y = this.Y
            };
        }

        public new Point3D Copy()
        {
            return new Point3D()
            {
                X = this.X,
                Y = this.Y,
                Z = this.Z
            };
        }

        #endregion
    }

    [Serializable]
    [TypeConverter(typeof(Point2DConverter))]
    public class Point2D
    {
        #region Properties

        [Category("Point"), Description("The x value of the point")]
        public double X { get; set; }

        [Category("Point"), Description("The y value of the point")]
        public double Y { get; set; }

        #endregion

        #region Setup

        public Point2D()
        {

        }

        public Point2D(double x, double y)
        {
            this.X = x;
            this.Y = y;
        }

        #endregion

        #region Operators

        public static Point2D operator -(Point2D left, Point2D right)
        {
            return new Point2D(left.X - right.X, left.Y - right.Y);
        }

        public static Point2D operator -(Point2D left, double right)
        {
            return new Point2D(left.X - right, left.Y - right);
        }

        public static Point2D operator +(Point2D left, Point2D right)
        {
            return new Point2D(left.X + right.X, left.Y + right.Y);
        }

        public static Point2D operator +(Point2D left, double right)
        {
            return new Point2D(left.X + right, left.Y + right);
        }

        public static Point2D operator *(Point2D left, double right)
        {
            return new Point2D(left.X * right, left.Y * right);
        }

        public static Point2D operator *(double left, Point2D right)
        {
            return new Point2D(left * right.X, left * right.Y);
        }

        public static Point2D operator /(Point2D left, double right)
        {
            return new Point2D(left.X / right, left.Y / right);
        }

        #endregion

        #region Methods

        public virtual double Length()
        {
            return Math.Sqrt(X * X + Y * Y);
        }

        public virtual double SquaredLength()
        {
            return X * X + Y * Y;
        }

        public double DistanceTo(Point2D point)
        {
            return (this - point).Length();
        }

        public double SquaredDistanceTo(Point2D point)
        {
            return (this - point).SquaredLength();
        }

        public virtual double MaxValue()
        {
            return Math.Max(X, Y);
        }

        public virtual double MinValue()
        {
            return Math.Min(X, Y);
        }

        public Point2D Copy()
        {
            return new Point2D()
            {
                X = this.X,
                Y = this.Y
            };
        }

        #endregion
    }

    [TypeConverter(typeof(CodedPointConverter))]
    public class CodedPoint : Point3D
    {
        #region Properties

        [Category("Style"), Description("The thickness of the point")]
        public double Thickness { get; set; }

        public System.Drawing.Color Color = System.Drawing.Color.Black;

        [Category("Color"), Description("The color of the point")]
        public System.Drawing.Color PointColor
        {
            get { return Color; }
        }

        [Category("Code"), Description("The assigned point code")]
        public string Code { get; set; }

        #endregion

        #region Setup

        public CodedPoint()
        {
            this.X = 0;
            this.Y = 0;
            this.Z = 0;
            this.Code = "";
            this.Thickness = 0.2f;
        }

        public CodedPoint(double x, double y, double z, byte r = 0, byte g = 0, byte b = 0, string code = "", float thickness = 0.2f)
        {
            this.X = x;
            this.Y = y;
            this.Z = z;
            this.Code = code;
            this.Thickness = thickness;
        }

        #endregion

        #region Methods

        public new CodedPoint Copy()
        {
            return new CodedPoint()
            {
                X = this.X,
                Y = this.Y,
                Z = this.Z,
                Color = this.Color,
                Code = this.Code,
                Thickness = this.Thickness,
                id = this.id
            };
        }

        public int id;

        public int ID
        {
            get { return id; }
        }

        public void SetID(int id)
        {
            this.id = id;
        }

        #endregion
    }

    public class SectionPoint : Point3D
    {
        #region Properties

        public double Brg;
        public byte R;
        public byte G;
        public byte B;

        #endregion

        #region Setup

        public SectionPoint()
        {
            X = 0;
            Y = 0;
            Z = 0;
            Brg = 0;
            R = 0;
            G = 0;
            B = 0;
        }

        public SectionPoint(double x, double y, double z, byte r, byte g, byte b)
        {
            this.X = x;
            this.Y = y;
            this.Z = z;
            Brg = 0;
            this.R = r;
            this.G = g;
            this.B = b;
        }

        #endregion

        #region Methods

        public new SectionPoint Copy()
        {
            return new SectionPoint()
            {
                X = this.X,
                Y = this.Y,
                Z = this.Z,
                Brg = this.Brg,
                R = this.R,
                G = this.G,
                B = this.B
            };
        }

        #endregion
    }

    public class Ray
    {
        #region Properties

        public Point3D Origin = new Point3D();
        public Point3D End = new Point3D();
        public Point3D Direction = new Point3D();

        #endregion

        #region Setup

        public Ray(Point3D origin, Point3D direction, Point3D end)
        {
            this.Origin = origin;
            this.Direction = direction;
            this.End = end;
        }

        public Ray()
        {

        }

        #endregion
    }

    [Serializable]
    public class CloudPoint : Point3D
    {
        #region Properties

        public byte R;
        public byte G;
        public byte B;
        public ushort Intensity = 0;

        #endregion

        #region Setup

        public CloudPoint()
        {

        }

        public CloudPoint(double x, double y, double z)
        {
            this.X = x;
            this.Y = y;
            this.Z = z;
        }

        public CloudPoint(double x, double y, double z, byte r, byte g, byte b)
        {
            this.X = x;
            this.Y = y;
            this.Z = z;
            this.R = r;
            this.G = g;
            this.B = b;
        }

        public CloudPoint(double x, double y, double z, byte r, byte g, byte b, ushort intensity)
        {
            this.X = x;
            this.Y = y;
            this.Z = z;
            this.R = r;
            this.G = g;
            this.B = b;
            this.Intensity = intensity;
        }

        #endregion

        #region Methods

        public new CloudPoint Copy()
        {
            return new CloudPoint()
            {
                X = this.X,
                Y = this.Y,
                Z = this.Z,
                R = this.R,
                G = this.G,
                B = this.B,
                Intensity = this.Intensity
            };

        }

        #endregion
    }

    [Serializable]
    public class Bounds
    {
        #region Properties

        public Point3D Min;
        public Point3D Max;
        public Point3D Size
        {
            get { return Max - Min; }
        }

        public Point3D Center
        {
            get { return (Max + Min) / 2; }
        }

        #endregion

        #region Setup

        public Bounds()
        {
            Min = new Point3D(double.MaxValue, double.MaxValue, double.MaxValue);
            Max = new Point3D(double.MinValue, double.MinValue, double.MinValue);
        }

        public Bounds(Point3D min, Point3D max)
        {
            Min = min;
            Max = max;
        }

        public Bounds(double minX, double minY, double minZ, double maxX, double maxY, double maxZ)
        {
            Min = new Point3D(minX, minY, minZ);
            Max = new Point3D(maxX, maxY, maxZ);
        }

        #endregion

        #region Methods

        public Bounds Copy()
        {
            var newBounds = new Bounds();

            if (this.Min != null)
                newBounds.Min = this.Min.Copy();
            if (this.Max != null)
                newBounds.Max = this.Max.Copy();

            return newBounds;
        }

        public bool IsInside(Point3D point)
        {
            if (Min.X <= point.X && point.X <= Max.X)
                if (Min.Y <= point.Y && point.Y <= Max.Y)
                    if (Min.Z <= point.Z && point.Z <= Max.Z)
                        return true;

            return false;
        }

        public bool IsInside(double x, double y, double z)
        {
            if (Min.X <= x && x <= Max.X)
                if (Min.Y <= y && y <= Max.Y)
                    if (Min.Z <= z && z <= Max.Z)
                        return true;
            return false;
        }

        public bool IsInside2D(double x, double y, double z)
        {
            if (Min.X <= x && x <= Max.X)
                if (Min.Y <= y && y <= Max.Y)
                    return true;

            return false;
        }

        public bool IsInside2D(Point2D point)
        {
            if (Min.X <= point.X && point.X <= Max.X)
                if (Min.Y <= point.Y && point.Y <= Max.Y)
                    return true;

            return false;
        }

        public bool IsInside2D(Point3D point)
        {
            if (Min.X <= point.X && point.X <= Max.X)
                if (Min.Y <= point.Y && point.Y <= Max.Y)
                    return true;

            return false;
        }

        public bool Intersects2D(Bounds b)
        {
            return (this.Min.X <= b.Max.X && this.Max.X >= b.Min.X) && (this.Min.Y <= b.Max.Y && this.Max.Y >= b.Min.Y);
        }

        public bool Intersects3D(Bounds b)
        {
            return (this.Min.X <= b.Max.X && this.Max.X >= b.Min.X) && (this.Min.Y <= b.Max.Y && this.Max.Y >= b.Min.Y) && (this.Min.Z <= b.Max.Z && this.Max.Z >= b.Min.Z);
        }

        public void Update(Point3D point)
        {
            Min.X = Math.Min(Min.X, point.X);
            Min.Y = Math.Min(Min.Y, point.Y);
            Min.Z = Math.Min(Min.Z, point.Z);

            Max.X = Math.Max(Max.X, point.X);
            Max.Y = Math.Max(Max.Y, point.Y);
            Max.Z = Math.Max(Max.Z, point.Z);
        }

        public void Update(Point2D point)
        {
            Min.X = Math.Min(Min.X, point.X);
            Min.Y = Math.Min(Min.Y, point.Y);

            Max.X = Math.Max(Max.X, point.X);
            Max.Y = Math.Max(Max.Y, point.Y);
        }

        public void Update(double x, double y)
        {
            Min.X = Math.Min(Min.X, x);
            Min.Y = Math.Min(Min.Y, y);

            Max.X = Math.Max(Max.X, x);
            Max.Y = Math.Max(Max.Y, y);
        }

        public void Update(double x, double y, double z)
        {
            Min.X = Math.Min(Min.X, x);
            Min.Y = Math.Min(Min.Y, y);
            Min.Z = Math.Min(Min.Z, z);

            Max.X = Math.Max(Max.X, x);
            Max.Y = Math.Max(Max.Y, y);
            Max.Z = Math.Max(Max.Z, z);
        }

        public void Update(Bounds bounds)
        {
            Update(bounds.Min);
            Update(bounds.Max);
        }

        public void MakeCubic()
        {
            Max = Min + Size.MaxValue();
        }

        public void SetBoundLines(ref float[] vertices, ref float[] colors, ref int count, float offsetX, float offsetY, float offsetZ)
        {
            count = 24;
            vertices = new float[72];
            colors = new float[72];
            int index = 0;

            //Back panel
            AddTopLeftDown(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddTopLeftUp(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddTopLeftDown(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddTopRightDown(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddTopLeftUp(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddTopRightUp(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddTopRightDown(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddTopRightUp(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);

            //Front panel
            AddBotLeftDown(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddBotLeftUp(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddBotLeftDown(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddBotRightDown(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddBotLeftUp(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddBotRightUp(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddBotRightDown(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddBotRightUp(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);

            //Left Side
            AddBotLeftDown(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddTopLeftDown(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddBotLeftUp(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddTopLeftUp(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);

            //Right side
            AddBotRightDown(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddTopRightDown(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddBotRightUp(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
            AddTopRightUp(ref vertices, ref colors, ref index, offsetX, offsetY, offsetZ);
        }

        private void AddTopLeftUp(ref float[] vertices, ref float[] colors, ref int count, float offsetX, float offsetY, float offsetZ)
        {
            vertices[count] = (float)Min.X - offsetX;
            vertices[count + 1] = (float)Max.Y - offsetY;
            vertices[count + 2] = (float)Max.Z - offsetZ;
            colors[count] = 1.0f;
            colors[count + 1] = 0.953f;
            colors[count + 2] = 0.0588f;
            count += 3;
        }

        private void AddTopLeftDown(ref float[] vertices, ref float[] colors, ref int count, float offsetX, float offsetY, float offsetZ)
        {
            vertices[count] = (float)Min.X - offsetX;
            vertices[count + 1] = (float)Max.Y - offsetY;
            vertices[count + 2] = (float)Min.Z - offsetZ;
            colors[count] = 1.0f;
            colors[count + 1] = 0.953f;
            colors[count + 2] = 0.0588f;
            count += 3;
        }

        private void AddBotLeftUp(ref float[] vertices, ref float[] colors, ref int count, float offsetX, float offsetY, float offsetZ)
        {
            vertices[count] = (float)Min.X - offsetX;
            vertices[count + 1] = (float)Min.Y - offsetY;
            vertices[count + 2] = (float)Max.Z - offsetZ;
            colors[count] = 1.0f;
            colors[count + 1] = 0.953f;
            colors[count + 2] = 0.0588f;
            count += 3;
        }

        private void AddBotLeftDown(ref float[] vertices, ref float[] colors, ref int count, float offsetX, float offsetY, float offsetZ)
        {
            vertices[count] = (float)Min.X - offsetX;
            vertices[count + 1] = (float)Min.Y - offsetY;
            vertices[count + 2] = (float)Min.Z - offsetZ;
            colors[count] = 1.0f;
            colors[count + 1] = 0.953f;
            colors[count + 2] = 0.0588f;
            count += 3;
        }

        private void AddTopRightUp(ref float[] vertices, ref float[] colors, ref int count, float offsetX, float offsetY, float offsetZ)
        {
            vertices[count] = (float)Max.X - offsetX;
            vertices[count + 1] = (float)Max.Y - offsetY;
            vertices[count + 2] = (float)Max.Z - offsetZ;
            colors[count] = 1.0f;
            colors[count + 1] = 0.953f;
            colors[count + 2] = 0.0588f;
            count += 3;
        }

        private void AddTopRightDown(ref float[] vertices, ref float[] colors, ref int count, float offsetX, float offsetY, float offsetZ)
        {
            vertices[count] = (float)Max.X - offsetX;
            vertices[count + 1] = (float)Max.Y - offsetY;
            vertices[count + 2] = (float)Min.Z - offsetZ;
            colors[count] = 1.0f;
            colors[count + 1] = 0.953f;
            colors[count + 2] = 0.0588f;
            count += 3;
        }

        private void AddBotRightUp(ref float[] vertices, ref float[] colors, ref int count, float offsetX, float offsetY, float offsetZ)
        {
            vertices[count] = (float)Max.X - offsetX;
            vertices[count + 1] = (float)Min.Y - offsetY;
            vertices[count + 2] = (float)Max.Z - offsetZ;
            colors[count] = 1.0f;
            colors[count + 1] = 0.953f;
            colors[count + 2] = 0.0588f;
            count += 3;
        }

        private void AddBotRightDown(ref float[] vertices, ref float[] colors, ref int count, float offsetX, float offsetY, float offsetZ)
        {
            vertices[count] = (float)Max.X - offsetX;
            vertices[count + 1] = (float)Min.Y - offsetY;
            vertices[count + 2] = (float)Min.Z - offsetZ;
            colors[count] = 1.0f;
            colors[count + 1] = 0.953f;
            colors[count + 2] = 0.0588f;
            count += 3;
        }

        public bool IntersectsWithRay(Models.Ray ray)
        {
            //Check x plane
            double tNear = double.NegativeInfinity, tFar = double.PositiveInfinity;
            if (ray.Direction.X == 0)
            {
                //Ray is parallel to x plane and must be between bounds
                if (ray.Origin.X < Min.X || ray.Origin.X > Max.X)
                    return false;
            }
            else
            {
                var t1 = (Min.X - ray.Origin.X) / ray.Direction.X;
                var t2 = (Max.X - ray.Origin.X) / ray.Direction.X;
                if (t1 > t2)
                    (t1, t2) = (t2, t1);
                if (t1 > tNear)
                    tNear = t1;
                if (t2 < tFar)
                    tFar = t2;
                if (tNear > tFar)
                    return false;
                if (tFar < 0)
                    return false;
            }

            //Check y plane
            if (ray.Direction.Y == 0)
            {
                //Ray is parallel to y plane and must be between bounds
                if (ray.Origin.Y < Min.Y || ray.Origin.Y > Max.Y)
                    return false;
            }
            else
            {
                var t1 = (Min.Y - ray.Origin.Y) / ray.Direction.Y;
                var t2 = (Max.Y - ray.Origin.Y) / ray.Direction.Y;
                if (t1 > t2)
                    (t1, t2) = (t2, t1);
                if (t1 > tNear)
                    tNear = t1;
                if (t2 < tFar)
                    tFar = t2;
                if (tNear > tFar)
                    return false;
                if (tFar < 0)
                    return false;
            }

            //Check z plane
            if (ray.Direction.Z == 0)
            {
                //Ray is parallel to z plane and must be between bounds
                if (ray.Origin.Z < Min.Z || ray.Origin.Z > Max.Z)
                    return false;
            }
            else
            {
                var t1 = (Min.Z - ray.Origin.Z) / ray.Direction.Z;
                var t2 = (Max.Z - ray.Origin.Z) / ray.Direction.Z;
                if (t1 > t2)
                    (t1, t2) = (t2, t1);
                if (t1 > tNear)
                    tNear = t1;
                if (t2 < tFar)
                    tFar = t2;
                if (tNear > tFar)
                    return false;
                if (tFar < 0)
                    return false;
            }

            return true;
        }

        public void GenerateLocalCorners(ref Point3D topLeft, ref Point3D topRight, ref Point3D botRight, ref Point3D botLeft, double zValue)
        {
            topLeft.X = Min.X - Center.X;
            topLeft.Y = Min.Y - Center.Y;
            topLeft.Z = zValue;

            topRight.X = Max.X - Center.X;
            topRight.Y = Min.Y - Center.Y;
            topRight.Z = zValue;

            botRight.X = Max.X - Center.X;
            botRight.Y = Max.Y - Center.Y;
            botRight.Z = zValue;

            botLeft.X = Min.Y - Center.X;
            botLeft.Y = Max.Y - Center.Y;
            botLeft.Z = zValue;
        }

        /// <summary>
        /// Generates a 2D array of corners coordinates in the order topLeft, topRight, botRight, botLeft
        /// </summary>
        /// <param name="zValue"></param>
        /// <returns></returns>
        public float[,] GenerateLocalCorners(float zValue)
        {
            var corners = new float[4, 3];
            corners[0, 0] = (float)(Min.X - Center.X);
            corners[0, 1] = (float)(Min.Y - Center.Y);
            corners[0, 2] = zValue;

            corners[1, 0] = (float)(Max.X - Center.X);
            corners[1, 1] = (float)(Min.Y - Center.Y);
            corners[1, 2] = zValue;

            corners[2, 0] = (float)(Max.X - Center.X);
            corners[2, 1] = (float)(Max.Y - Center.Y);
            corners[2, 2] = zValue;

            corners[3, 0] = (float)(Min.X - Center.X);
            corners[3, 1] = (float)(Max.Y - Center.Y);
            corners[3, 2] = zValue;

            return corners;
        }

        public double CalcSpacingFromDistance(double originalDist, int originalWidth, Point3D cameraPos)
        {
            var cX = Min.X + (Size.X / 2);
            var cY = Min.Y + (Size.Y / 2);
            var cZ = Min.Z + (Size.Z / 2);

            var distToCenter = Vector.FnDistance(cameraPos.X, cameraPos.Y, cX, cY);

            var newDist = originalDist + (distToCenter);
            return newDist / originalWidth;
        }

        public bool FindClosestIntersection2D(Point3D lineStart, Point3D lineEnd, ref Point3D intersectPoint)
        {
            var boundPoints = new List<Point3D> { new Point3D(Min.X, Min.Y, 0), new Point3D(Min.X, Max.Y, 0), new Point3D(Max.X, Max.Y, 0), new Point3D(Max.X, Min.Y, 0), new Point3D(Min.X, Min.Y, 0) };

            double dist = double.MaxValue;
            bool doesIntersect = false;

            for(int i = 0; i < boundPoints.Count - 1; i++)
            {
                var segStart = boundPoints[i];
                var segEnd = boundPoints[i + 1];
                

                double ipX = 0, ipY = 0;
                if(Intersection.IntersectLines(lineStart.X, lineStart.Y, lineEnd.X, lineEnd.Y, segStart.X, segStart.Y, segEnd.X, segEnd.Y, ref ipX, ref ipY))
                {
                    //Intersection is found so test it is valid
                    if(Intersection.FnInBox1(segStart.X, segStart.Y, segEnd.X, segEnd.Y, ipX, ipY))
                    {
                        var distToIntersection = Vector.FnDistance(lineStart.X, lineStart.Y, ipX, ipY);
                        if(distToIntersection < dist && distToIntersection > 0.0001)
                        {
                            doesIntersect = true;
                            dist = distToIntersection;
                            intersectPoint = new Point3D(ipX, ipY, 0);
                        }
                    }
                }
            }

            return doesIntersect;
        }

        #endregion
    }
}
