﻿using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using NRG.Services;
using NRG.Import;
using NRG.Export;
using NRG.Maths;

namespace NRG.Models
{
    public class Converter
    {
        #region Properties

        private List<string> sources;
        private List<AsciiFormat> sourceFormats;
        private string workDir;
        private string projectName = "";
        private string projectPath = "";
        private double spacing;
        private List<double> colorRange = new List<double>();
        private List<double> intensityRange = new List<double>();
        private double scale = 0;
        private int diagonalFraction = 200;
        private List<ulong> pointsList = new List<ulong>();

        #endregion

        #region Setup

        public Converter(string workDir, string projectName, string projectPath, List<string> sources)
        {
            this.workDir = workDir;
            this.projectName = projectName;
            this.projectPath = projectPath;
            this.sources = sources;
        }

        public Converter(string workDir, string projectName, string projectPath, List<string> sources, List<AsciiFormat> formats)
        {
            this.workDir = workDir;
            this.projectName = projectName;
            this.projectPath = projectPath;
            this.sources = sources;
            this.sourceFormats = formats;
        }

        #endregion

        #region Methods

        public void Convert(BackgroundWorker worker, ref DoWorkEventArgs e)
        {
            Prepare();

            var state = new ConversionProgress();
            state.Stage = ConversionStage.GetBounds;
            worker.ReportProgress(0, state);

            ulong pointsProcessed = 0;
            ulong NextProcessCheck = 10000000;
            
            var bounds = CalculateBounds(worker, ref e);
            if(worker.CancellationPending)
            {
                e.Cancel = true;
                return;
            }

            if (scale == 0)
            {
                //if (bounds.Size.Length() > 1000000)
                //    scale = 0.01;
                //else if (bounds.Size.Length() > 100000)
                //    scale = 0.001;
                //else if (bounds.Size.Length() > 1)
                //    scale = 0.001;
                //else
                    scale = 0.0001;
            }

            bounds.MakeCubic();

            if (diagonalFraction != 0)
                spacing = bounds.Size.Length() / diagonalFraction;

            Octree octree = null;
            octree = new Octree(this.workDir, this.projectName, this.projectPath, bounds, spacing, -1, scale);
            octree.Version = OctreeFileVersion.FileVersion2;

            ulong totalPoints = 0;
            foreach (var count in pointsList)
                totalPoints += count;

            if (octree == null)
                return;

            //Process the files
            pointsProcessed = 0;
            foreach (var source in sources)
            {
                ulong localPointsProcessed = 0;
                var numPoints = pointsList[sources.IndexOf(source)];

                //Report conversion process
                state.Stage = ConversionStage.BuildingOctree;
                state.PrimaryPercentage = (int)(100 * ((double)sources.IndexOf(source) / (double)sources.Count));
                state.SecondaryPercentrage = 0;
                state.SecondaryMessage = Path.GetFileName(source);
                state.TotalPoints = totalPoints;
                state.CurrentPoints = pointsProcessed;
                state.TotalFilePoints = numPoints;
                state.currentFilePoints = localPointsProcessed;
                worker.ReportProgress(0, state);

                PointReader reader = null;

                if (sourceFormats != null && sourceFormats.Count >= 0)
                    reader = CreatePointReader(source, sourceFormats[sources.IndexOf(source)]);
                else
                    reader = CreatePointReader(source);

                double fraction = 1.0D / (double)numPoints;
                var lastProgress = 0;

                CloudPoint p = null;

                while (reader.ReadNextPoint(ref p))
                {
                    localPointsProcessed++;
                    pointsProcessed++;

                    octree.Add(p);
                    if ((int)(localPointsProcessed * fraction * 100) > lastProgress)
                    {
                        if(worker.CancellationPending)
                        {
                            octree.WaitUntillProcessed();
                            reader.Close();
                            reader = null;
                            e.Cancel = true;
                            return;
                        }

                        lastProgress = (int)(localPointsProcessed * fraction * 100);
                        state.CurrentPoints = pointsProcessed;
                        state.currentFilePoints = localPointsProcessed;
                        state.SecondaryPercentrage = lastProgress;
                        worker.ReportProgress(0, state);
                    }

                    if (pointsProcessed == NextProcessCheck)
                    {
                        if (worker.CancellationPending)
                        {
                            octree.WaitUntillProcessed();
                            reader.Close();
                            reader = null;
                            e.Cancel = true;
                            return;
                        }

                        NextProcessCheck += 10000000;
                        octree.WaitUntillProcessed();
                        state.Stage = ConversionStage.FlushingOctree;
                        worker.ReportProgress(0, state);
                        octree.Flush();
                        state.Stage = ConversionStage.BuildingOctree;
                        worker.ReportProgress(0, state);
                    }
                }

                reader.Close();
                reader = null;
            }

            state.Stage = ConversionStage.FinalizingOctree;
            state.CurrentPoints = pointsProcessed;
            worker.ReportProgress(0, state);
            octree.WaitUntillProcessed();
            octree.Flush();
        }

        public void MergePointCloud(BackgroundWorker worker, Octree currentOctree, ref DoWorkEventArgs e)
        {
            Prepare();

            var state = new ConversionProgress();
            state.Stage = ConversionStage.GetBounds;
            worker.ReportProgress(0, state);

            ulong pointsProcessed = 0;
            ulong nextProcessCheck = 10000000;
            var bounds = CalculateBounds(worker, ref e);

            //Add point cloud bounds
            bounds.Update(currentOctree.TightBounds);

            if(scale == 0)
            {
                if (bounds.Size.Length() > 1000000)
                    scale = 0.01;
                else if (bounds.Size.Length() > 100000)
                    scale = 0.001;
                else if (bounds.Size.Length() > 1)
                    scale = 0.001;
                else
                    scale = 0.0001;
            }

            bounds.MakeCubic();

            if (diagonalFraction != 0)
                spacing = bounds.Size.Length() / diagonalFraction;

            Octree octree = null;
            octree = new Octree(this.workDir, this.projectName, this.projectPath, bounds, spacing, -1, scale);
            octree.Version = OctreeFileVersion.FileVersion2;

            if (octree == null)
                return;

            //Process the point cloud
            var nodes = new List<OctreeNode>();
            currentOctree.Root.GetAllNodes(ref nodes);
            ulong totalPoints = 0;
            foreach (var count in pointsList)
                totalPoints += count;

            totalPoints += currentOctree.NumAccepted;

            //Report conversion progress
            state.Stage = ConversionStage.BuildingOctree;
            state.PrimaryPercentage = 0;
            state.SecondaryPercentrage = 0;
            state.SecondaryMessage = Path.GetFileNameWithoutExtension(currentOctree.ProjectName);
            state.TotalPoints = totalPoints;
            state.CurrentPoints = pointsProcessed;
            state.TotalFilePoints = currentOctree.NumAccepted;
            state.currentFilePoints = 0;
            worker.ReportProgress(0, state);

            ulong localPointsProcessed = 0;
            var numPoints = currentOctree.NumAccepted;
            double fraction = 1.0D / numPoints;

            foreach (var node in nodes)
            {
                var points = node.LoadIntoStore();

                var lastProgress = 0;

                foreach(var point in points)
                {
                    localPointsProcessed++;
                    pointsProcessed++;

                    octree.Add(point);
                    if ((int)(localPointsProcessed * fraction * 100) > lastProgress)
                    {
                        if(worker.CancellationPending)
                        {
                            octree.WaitUntillProcessed();
                            e.Cancel = true;
                            return;
                        }

                        lastProgress = (int)(localPointsProcessed * fraction * 100);
                        state.CurrentPoints = pointsProcessed;
                        state.currentFilePoints = localPointsProcessed;
                        state.SecondaryPercentrage = lastProgress;
                        worker.ReportProgress(0, state);
                    }

                    if(pointsProcessed == nextProcessCheck)
                    {
                        if(worker.CancellationPending)
                        {
                            octree.WaitUntillProcessed();
                            e.Cancel = true;
                            return;
                        }

                        nextProcessCheck += 10000000;
                        octree.WaitUntillProcessed();
                        state.Stage = ConversionStage.FlushingOctree;
                        worker.ReportProgress(0, state);
                        octree.Flush();
                        state.Stage = ConversionStage.BuildingOctree;
                    }
                }
            }

            //Process the files
            foreach(var source in sources)
            {
                localPointsProcessed = 0;
                numPoints = pointsList[sources.IndexOf(source)];

                state.Stage = ConversionStage.BuildingOctree;
                state.PrimaryPercentage = (int)(100 * (((double)sources.IndexOf(source) + 1) / ((double)sources.Count + 1)));
                state.SecondaryPercentrage = 0;
                state.SecondaryMessage = Path.GetFileNameWithoutExtension(source);
                state.TotalPoints = totalPoints;
                state.CurrentPoints = pointsProcessed;
                state.TotalFilePoints = currentOctree.NumAccepted;
                state.currentFilePoints = 0;
                worker.ReportProgress(0, state);

                PointReader reader = null;

                if (sourceFormats != null && sourceFormats.Count >= 0)
                    reader = CreatePointReader(source, sourceFormats[sources.IndexOf(source)]);
                else
                    reader = CreatePointReader(source);

                fraction = 1.0D / (double)numPoints;
                var lastProgress = 0;

                CloudPoint p = null;

                while(reader.ReadNextPoint(ref p))
                {
                    localPointsProcessed++;
                    pointsProcessed++;

                    octree.Add(p);
                    if((int)(localPointsProcessed * fraction * 100) > lastProgress)
                    {
                        if (worker.CancellationPending)
                        {
                            octree.WaitUntillProcessed();
                            reader.Close();
                            reader = null;
                            e.Cancel = true;
                            return;
                        }

                        lastProgress = (int)(localPointsProcessed * fraction * 100);
                        state.CurrentPoints = pointsProcessed;
                        state.currentFilePoints = localPointsProcessed;
                        state.SecondaryPercentrage = lastProgress;
                        worker.ReportProgress(0, state);
                    }

                    if(pointsProcessed == nextProcessCheck)
                    {
                        if (worker.CancellationPending)
                        {
                            octree.WaitUntillProcessed();
                            reader.Close();
                            reader = null;
                            e.Cancel = true;
                            return;
                        }

                        nextProcessCheck += 10000000;
                        octree.WaitUntillProcessed();
                        state.Stage = ConversionStage.FlushingOctree;
                        worker.ReportProgress(0, state);
                        octree.Flush();
                        state.Stage = ConversionStage.BuildingOctree;
                    }
                }

                reader.Close();
                reader = null;
            }

            state.Stage = ConversionStage.FinalizingOctree;
            state.CurrentPoints = pointsProcessed;
            worker.ReportProgress(0, state);
            octree.WaitUntillProcessed();
            octree.Flush();
        }

        private PointReader CreatePointReader(string source, AsciiFormat format = null)
        {
            PointReader reader = null;
            var upperSource = source.ToUpper();

            if (upperSource.EndsWith(".LAS") || upperSource.EndsWith(".LAZ"))
                reader = new LASPointReader(source);
            else if(upperSource.EndsWith(".XYZ") || upperSource.EndsWith(".TXT") || upperSource.EndsWith(".PTX") || upperSource.EndsWith(".PTY") || upperSource.EndsWith(".CSV") || upperSource.EndsWith(".GPF"))
            {
                if(format != null)
                {
                    reader = new AsciiReader(source, format, colorRange, intensityRange);
                }
            }
            else if(upperSource.EndsWith(".PTS"))
            {
                if(format != null)
                {
                    intensityRange = new List<double>();
                    intensityRange.Add(-2048);
                    intensityRange.Add(+2047);
                    reader = new AsciiReader(source, format, colorRange, intensityRange);
                }
            }
            else if(upperSource.EndsWith(".BIN"))
            {
                reader = new BinPointReader(source);
            }
            return reader;
        }

        private void Prepare()
        {
            //If sources contain directories use files inside the directory instead
            var sourceFiles = new List<string>();
            foreach(var source in sources)
            {
                if(Directory.Exists(source))
                {
                    //Its a directory
                    foreach(var file in Directory.GetFiles(source))
                    {
                        //Its a file
                        var upperFile = file.ToUpper();
                        if (upperFile.EndsWith(".LAS") || upperFile.EndsWith(".LAZ") || upperFile.EndsWith(".XYZ") || upperFile.EndsWith(".PTX") || upperFile.EndsWith(".TXT") || upperFile.EndsWith(".CSV") || upperFile.EndsWith(".GPF") || upperFile.EndsWith(".PTS"))
                            sourceFiles.Add(file);
                    }
                }
                else
                {
                    //Its a file
                    var upperFile = source.ToUpper();
                    if(upperFile.EndsWith(".LAS") || upperFile.EndsWith(".LAZ") || upperFile.EndsWith(".XYZ") || upperFile.EndsWith(".PTX") || upperFile.EndsWith(".TXT") || upperFile.EndsWith(".CSV") || upperFile.EndsWith(".GPF") || upperFile.EndsWith(".PTS") || upperFile.EndsWith(".BIN"))
                        sourceFiles.Add(source);
                }
            }

            this.sources = sourceFiles;
        }

        private Bounds CalculateBounds(BackgroundWorker worker, ref DoWorkEventArgs e)
        {
            Bounds newBounds = new Bounds();
            foreach(var source in sources)
            {
                if(worker.CancellationPending)
                {
                    e.Cancel = true;
                    return newBounds;
                }

                PointReader reader = null;
                if (sourceFormats != null && sourceFormats.Count >= 0)
                    reader = CreatePointReader(source, sourceFormats[sources.IndexOf(source)]);
                else
                    reader = CreatePointReader(source);

                Bounds lBounds = reader.GetBounds();
                pointsList.Add(reader.NumPoints());
                newBounds.Update(lBounds);

                reader.Close();
                reader = null;
            }

            return newBounds;
        }

        #endregion
    }

    public class Octree
    {
        #region Properties

        public long NumPoints;
        public Bounds Bounds = new Bounds();
        public Bounds TightBounds = new Bounds();
        public Bounds OffsetBounds = new Bounds();
        public string WorkDir;
        public double Spacing;
        public double Scale;
        public int MaxDepth = -1;
        public OctreeNode Root;
        public ulong NumAdded = 0;
        public ulong NumAccepted = 0;
        public string ProjectName;
        public string ProjectPath;
        public int HierachyStepSize = 5;
        public List<CloudPoint> Store = new List<CloudPoint>(1000000);
        public List<OctreeNode> LoadedNodes = new List<OctreeNode>();
        public List<OctreeNode> SelectedLoadedNodes = new List<OctreeNode>();
        public Thread StoreThread;
        public int PointsInMemory = 0;
        public OctreeFileVersion Version;
        public IntensityConverter IntensityConverter;

        public string LocalWorkDir
        {
            get
            {
                string localPath = "";

                var projectDir = Path.GetDirectoryName(ProjectPath);
                localPath = WorkDir.Replace(projectDir, "");

                return localPath;
            }
        }

        #endregion

        #region Setup

        public Octree(string projectPath)
        {
            this.ProjectPath = projectPath;
            if (ProjectPath.ToUpper().EndsWith(".JS") || ProjectPath.ToUpper().EndsWith(".PC"))
                Version = OctreeFileVersion.FileVersion1;
            else
                Version = OctreeFileVersion.FileVersion2;
        }

        public Octree(string workDir, string projectName, string projectPath)
        {
            this.WorkDir = workDir;
            this.ProjectName = projectName;
            this.ProjectPath = projectPath;

            if (ProjectPath.ToUpper().EndsWith(".JS") || ProjectPath.ToUpper().EndsWith(".PC"))
                Version = OctreeFileVersion.FileVersion1;
            else
                Version = OctreeFileVersion.FileVersion2;
        }

        public Octree(string workDir, string projectName, string projectPath, Bounds bounds, double spacing, int maxDepth, double scale)
        {
            this.WorkDir = workDir;
            this.ProjectName = projectName;
            this.ProjectPath = projectPath;
            this.Bounds = bounds;
            this.Spacing = spacing;
            this.MaxDepth = maxDepth;
            this.Scale = scale;

            if (ProjectPath.ToUpper().EndsWith(".JS") || ProjectPath.ToUpper().EndsWith(".PC"))
                Version = OctreeFileVersion.FileVersion1;
            else
                Version = OctreeFileVersion.FileVersion2;

            if(this.Scale == 0)
            {
                if (bounds.Size.Length() > 1000000)
                    this.Scale = 0.01;
                else if (bounds.Size.Length() > 100000)
                    this.Scale = 0.001;
                else if (bounds.Size.Length() > 1)
                    this.Scale = 0.001;
                else
                    this.Scale = 0.0001;
            }

            var scaleFactor = 1 / this.Scale;

            OffsetBounds = new Bounds();
            OffsetBounds.Update(Math.Round(bounds.Min.X / scaleFactor) * scaleFactor, Math.Round(bounds.Min.Y / scaleFactor) * scaleFactor, 0);

            Root = new OctreeNode(this, bounds);
        }

        #endregion

        #region Methods

        public void Add(CloudPoint point)
        {
            if (NumAdded == 0)
            {
                var dataDir = WorkDir + "/data";
                var tempDir = WorkDir + "/temp";

                Directory.CreateDirectory(dataDir);
                Directory.CreateDirectory(tempDir);
            }

            Store.Add(point);
            NumAdded++;

            if (Store.Count >= 1000000)
                ProcessStore();
        }

        public void ProcessStore()
        {
            List<CloudPoint> st = Store;
            Store = new List<CloudPoint>();

            WaitUntillProcessed();
            StoreThread = new Thread(new ParameterizedThreadStart(Process));
            StoreThread.Start(st);
        }

        private void Process(object state)
        {
            List<CloudPoint> st = (List<CloudPoint>)state;
            foreach (var p in st)
            {
                if (Root.Add(p))
                {
                    PointsInMemory++;
                    NumAccepted++;
                }
            }
        }

        public void WaitUntillProcessed()
        {
            if (StoreThread != null)
            {
                if (StoreThread.ThreadState == ThreadState.Running)
                    StoreThread.Join();
            }
        }

        public void Flush()
        {
            ProcessStore();

            if (StoreThread != null)
            {
                if (StoreThread.ThreadState == ThreadState.Running)
                    StoreThread.Join();
            }

            //Get all nodes to Flush
            var nodesToFlush = new List<OctreeNode>();
            Root.GetAllNodes(ref nodesToFlush);

            Parallel.ForEach(nodesToFlush, new ParallelOptions() { MaxDegreeOfParallelism = 4 }, (node) =>
            {
                node.Flush();
            });

            GC.Collect();
            //Root.Flush();
            //TightBounds = Root.AcceptedBounds;

            //Save the document 

            SaveFile();
        }

        public void SaveFile()
        {
            if (File.Exists(ProjectPath))
                File.Delete(ProjectPath);

            var writer = new BinaryWriter(File.Open(ProjectPath, FileMode.Create, FileAccess.ReadWrite));
            writer.Write("NRG Point Cloud File Version 2");

            writer.Write(LocalWorkDir);
            writer.Write(ProjectName);
            writer.Write(NumAccepted);

            writer.Write(HierachyStepSize);
            writer.Write(Spacing);
            writer.Write(Scale);

            writer.Write(Bounds.Min.X);
            writer.Write(Bounds.Min.Y);
            writer.Write(Bounds.Min.Z);
            writer.Write(Bounds.Max.X);
            writer.Write(Bounds.Max.Y);
            writer.Write(Bounds.Max.Z);

            writer.Write(TightBounds.Min.X);
            writer.Write(TightBounds.Min.Y);
            writer.Write(TightBounds.Min.Z);
            writer.Write(TightBounds.Max.X);
            writer.Write(TightBounds.Max.Y);
            writer.Write(TightBounds.Max.Z);

            writer.Write(OffsetBounds.Min.X);
            writer.Write(OffsetBounds.Min.Y);
            writer.Write(OffsetBounds.Min.Z);

            SaveNodeInfo(Root, writer);
            writer.Close();
            writer = null;
        }

        public void LoadStateFromDisk()
        {
            switch (Version)
            {
                case OctreeFileVersion.FileVersion1:
                    LoadFileVersion1();
                    break;
                case OctreeFileVersion.FileVersion2:
                    LoadFileVersion2();
                    break;
            }
        }

        private void LoadFileVersion1()
        {
            //Project file
            var projectFile = ((Services.OldFileVersions.PointCloudFile)Helpers.ReadPointCloudFile(ProjectPath)).ToNewPointCloudFile();

            this.HierachyStepSize = projectFile.HierarchyStepSize;
            this.Spacing = projectFile.Spacing;
            this.Scale = projectFile.Scale;
            this.Bounds = projectFile.BoundingBox;
            this.TightBounds = projectFile.TightBoundingBox;
            this.NumAccepted = projectFile.NumAccepted;
            this.WorkDir = Path.GetDirectoryName(ProjectPath) + Path.GetFileNameWithoutExtension(ProjectPath);

            //Tree
            var hrcPaths = new List<string>();
            hrcPaths.Add(WorkDir + "/data/r/r.hrc");
            string rootDir = WorkDir + "/data/r/";
            foreach (var directory in Directory.GetDirectories(rootDir))
            {
                foreach (var file in Directory.GetFiles(directory))
                {
                    if (file.EndsWith(".hrc"))
                        hrcPaths.Add(file);
                }

                foreach (var dir in Directory.GetDirectories(directory))
                {
                    foreach (var file in Directory.GetFiles(dir))
                    {
                        if (file.EndsWith(".hrc"))
                            hrcPaths.Add(file);
                    }
                }
            }

            hrcPaths.Sort();
            hrcPaths.Reverse();

            OctreeNode root = new OctreeNode(this, projectFile.BoundingBox);
            foreach (var hrcPath in hrcPaths)
            {
                string hrcName = Path.GetFileNameWithoutExtension(hrcPath);
                OctreeNode hrcRoot = root.FindNode(hrcName);

                OctreeNode current = hrcRoot;
                current.AddedSinceLastFlush = false;
                current.IsInMemory = false;
                var nodes = new List<OctreeNode>();
                nodes.Add(hrcRoot);

                var fin = new BinaryReader(File.Open(hrcPath, FileMode.Open, FileAccess.Read, FileShare.Read));

                //To do complete
                var length = fin.BaseStream.Length;
                var bytes = fin.ReadBytes((int)length);
                for (int i = 0; 5 * i < bytes.Length; i++)
                {
                    current = nodes[i];
                    byte children = bytes[i * 5];
                    byte[] p = new byte[4] { bytes[(i * 5) + 1], bytes[(i * 5) + 2], bytes[(i * 5) + 3], bytes[(i * 5) + 4] };
                    uint ip = BitConverter.ToUInt32(p, 0);
                    uint numPoints = ip;

                    //W.A. 10/01/2019 there is something seriously wrong here, will investigate later, for now its a hack
                    if (current.NumAccepted < numPoints)
                        current.NumAccepted = numPoints;

                    if (children != 0)
                    {
                        current.Children = new OctreeNode[8];
                        for (int j = 0; j < 8; j++)
                        {
                            if ((children & (1 << j)) != 0)
                            {
                                Bounds cBounds = OctreeInterop.ChildBounds(current.Bounds, j);
                                OctreeNode child = new OctreeNode(this, j, cBounds, current.Level + 1);
                                child.Parent = current;
                                child.AddedSinceLastFlush = false;
                                child.IsInMemory = false;
                                current.Children[j] = child;
                                nodes.Add(child);
                            }
                        }
                    }
                }
            }
            this.Root = root;
            this.NumAdded = 1;
        }

        private void LoadFileVersion2()
        {
            var reader = new BinaryReader(File.Open(ProjectPath, FileMode.Open, FileAccess.Read, FileShare.Read));

            var version = reader.ReadString();
            if (version != "NRG Point Cloud File Version 2")
            {
                //Invalid file
                throw new Exception("Invalid file format");
            }
            ProjectName = Path.GetFileName(ProjectPath);
            //this.WorkDir = reader.ReadString();
            var dirString = reader.ReadString();
            if (Directory.Exists(dirString))
                this.WorkDir = dirString;
            else
                this.WorkDir = Path.GetDirectoryName(ProjectPath) + "\\" + Path.GetFileNameWithoutExtension(ProjectName)/* dirString*/;

            reader.ReadString();
            //this.ProjectName = reader.ReadString();
            this.NumAccepted = reader.ReadUInt64();
            this.HierachyStepSize = reader.ReadInt32();
            this.Spacing = reader.ReadDouble();
            this.Scale = reader.ReadDouble();

            this.Bounds = new Bounds();
            this.Bounds.Update(reader.ReadDouble(), reader.ReadDouble(), reader.ReadDouble());
            this.Bounds.Update(reader.ReadDouble(), reader.ReadDouble(), reader.ReadDouble());

            this.TightBounds = new Bounds();
            this.TightBounds.Update(reader.ReadDouble(), reader.ReadDouble(), reader.ReadDouble());
            this.TightBounds.Update(reader.ReadDouble(), reader.ReadDouble(), reader.ReadDouble());

            this.OffsetBounds = new Bounds();
            this.OffsetBounds.Update(reader.ReadDouble(), reader.ReadDouble(), reader.ReadDouble());

            Root = new OctreeNode(this, Bounds);
            LoadNodeInfo(Root, reader);

            reader.Close();
            reader = null;
        }

        private void LoadNodeInfo(OctreeNode node, BinaryReader reader)
        {
            //Read number of points
            var p = reader.ReadUInt32();
            node.NumAccepted = p;


            //Read children
            var children = reader.ReadByte();
            if (children != 0)
            {
                node.Children = new OctreeNode[8];
                for (int j = 0; j < 8; j++)
                {
                    if ((children & (1 << j)) != 0)
                    {
                        Bounds cBounds = OctreeInterop.ChildBounds(node.Bounds, j);
                        OctreeNode child = new OctreeNode(this, j, cBounds, node.Level + 1);
                        child.Parent = node;
                        child.AddedSinceLastFlush = false;
                        child.IsInMemory = false;
                        node.Children[j] = child;
                        LoadNodeInfo(child, reader);
                    }
                }
            }
            else
                return;
        }

        private void SaveNodeInfo(OctreeNode node, BinaryWriter writer)
        {
            //Save number of points
            writer.Write(node.NumAccepted);

            //Write children
            byte children = 0;
            if (node.Children != null)
            {
                for (int j = 0; j < (int)node.Children.Length; j++)
                {
                    if (node.Children[j] != null)
                        children = (byte)(children | (1 << j));
                }
            }
            writer.Write(children);

            if (node.Children != null)
            {
                for (int i = 0; i < node.Children.Length; i++)
                {
                    if (node.Children[i] != null)
                        SaveNodeInfo(node.Children[i], writer);
                }
            }
        }

        #endregion
    }

    public class OctreeNode
    {
        #region Properties

        public int Index = -1;
        public Bounds Bounds;
        public int Level = 0;
        public uint NumAccepted = 0;
        public Dictionary<(int, int, int), CloudPoint> Grid;
        public OctreeNode Parent = null;
        public OctreeNode[] Children;
        public bool AddedSinceLastFlush = true;
        public bool AddCalledSinceLastFlush = false;
        public Octree Octree;
        public int StoreLimit = 20000;
        public List<CloudPoint> Store = new List<CloudPoint>();
        public bool IsInMemory = true;
        public bool IsRendered = false;
        public bool IsSelectedRendered = false;
        public double GridSpacing;
        public double GridHalfSpacing;
        public double GridSize;
        public double SpacingFactor;
        public float[] vertices;
        public float[] colors;

        #endregion

        #region Setup

        public OctreeNode(Octree octree, Bounds bounds)
        {
            this.Octree = octree;
            this.Bounds = bounds;
            this.SpacingFactor = 1 / (Octree.Spacing / Math.Pow(2.0, (float)Level));
            SetupGrid();
        }

        public OctreeNode(Octree octree, int index, Bounds bounds, int level)
        {
            this.Index = index;
            this.Bounds = bounds;
            this.Level = level;
            this.Octree = octree;
            this.SpacingFactor = 1 / (Octree.Spacing / Math.Pow(2.0, (float)Level));
            SetupGrid();
        }

        #endregion

        #region Methods

        public string Name
        {
            get
            {
                if (Parent == null)
                    return "r";
                else
                    return Parent.Name + Index.ToString();
            }
        }

        public string FilePath
        {
            get
            {
                if (Octree.Version == OctreeFileVersion.FileVersion1)
                {
                    if (Parent == null)
                        return "r";
                    else
                    {
                        var name = Name;
                        if (name.Length >= Octree.HierachyStepSize + 1)
                        {
                            var splitName = "";
                            for (int i = 0; i < Math.Floor((double)name.Length / (double)Octree.HierachyStepSize); i++)
                            {
                                int end = 0;

                                if ((i * Octree.HierachyStepSize) + 1 + Octree.HierachyStepSize > name.Length)
                                    end = name.Length - ((i * Octree.HierachyStepSize) + 1);
                                else
                                    end = Octree.HierachyStepSize;

                                var substring = name.Substring((i * Octree.HierachyStepSize) + 1, Octree.HierachyStepSize);
                                splitName += substring + "\\";
                            }
                            splitName += name;
                            return splitName;
                        }
                        else
                            return name;
                    }
                }
                else if (Octree.Version == OctreeFileVersion.FileVersion2)
                {
                    if (Parent == null)
                        return "r";
                    else
                    {
                        var name = Name;
                        if (name.Length >= Octree.HierachyStepSize + 1)
                        {
                            var splitName = "";
                            for(int i = 0; i < Math.Floor(((double)name.Length - 1) / (double)Octree.HierachyStepSize); i++)
                            {
                                int end = 0;

                                if ((i * Octree.HierachyStepSize) + 1 + Octree.HierachyStepSize > name.Length)
                                    end = name.Length - ((i * Octree.HierachyStepSize) + 1);
                                else
                                    end = Octree.HierachyStepSize;

                                var substring = name.Substring((i * Octree.HierachyStepSize) + 1, end);
                                splitName += substring + "\\";
                            }

                            //Get remainder of the name
                            var characterCount = Level % 5;
                            if (characterCount == 0)
                                splitName += "r";
                            else
                                splitName += name.Substring(name.Length - characterCount, characterCount);
                          
                            return splitName;
                        }
                        else
                            return name;
                    }
                }
                else
                    return null;
            }
        }

        public double Spacing
        {
            get
            {
                return (double)(Octree.Spacing / Math.Pow(2.0, (float)Level));
            }
        }

        public string WorkDir
        {
            get
            {
                return Octree.WorkDir;
            }
        }

        public string HierarchyPath
        {
            get
            {
                string path = "r/";

                int hierarchyStepSize = Octree.HierachyStepSize;
                string indices = Name.Substring(1);

                int numParts = (int)Math.Floor((double)indices.Length / (float)hierarchyStepSize);
                for (int i = 0; i < numParts; i++)
                    path += indices.Substring(i * hierarchyStepSize, hierarchyStepSize) + "/";

                return path;
            }
        }

        public bool IsLeafNode
        {
            get
            {
                if (Children == null)
                    return true;
                else
                    return false;
            }
        }

        public bool IsInnerNode
        {
            get
            {
                if (Children.Length > 0)
                    return true;
                else
                    return false;
            }
        }

        public void LoadFromDisk()
        {
            var reader = new BinaryReader(File.Open(WorkDir + "/data/r/" + FilePath + ".bin", FileMode.Open, FileAccess.Read, FileShare.Read));
            var scale = Octree.Scale;
            var bounds = OffsetBounds;
            uint c = 0;
            uint totalPoints = (uint)(reader.BaseStream.Length / 18);

            //Seperating these is a minor optimization but consindering the size of the data we could save a few billion if statements
            //So its worth the extra code...
            if(IsLeafNode)
            {
                //Add to store
                while(c <totalPoints)
                {
                    c++;
                    try
                    {
                        var p = new CloudPoint();
                        p.X = ((reader.ReadInt32() * scale) + bounds.Min.X);
                        p.Y = ((reader.ReadInt32() * scale) + bounds.Min.Y);
                        p.Z = ((reader.ReadInt32() * scale) + bounds.Min.Z);

                        //Alpha value is stored but not used so we must take next 4 bytes not 3
                        byte[] rgba = reader.ReadBytes(4);
                        p.R = rgba[0];
                        p.G = rgba[1];
                        p.B = rgba[2];

                        p.Intensity = reader.ReadUInt16();

                        //Add direct to store since we already established this is a leaf node
                        Store.Add(p);
                    }
                    catch
                    {
                        //We can assume if it fails to read one point the data is likely corrupted
                        break;
                    }
                }
            }
            else
            {
                //Add to grid
                while(c < totalPoints)
                {
                    c++;
                    try
                    {
                        var p = new CloudPoint();
                        p.X = ((reader.ReadInt32() * scale) + bounds.Min.X);
                        p.Y = ((reader.ReadInt32() * scale) + bounds.Min.Y);
                        p.Z = ((reader.ReadInt32() * scale) + bounds.Min.Z);

                        //Alpha value is stored but not used so we must take next 4 bytes not 3
                        byte[] rgba = reader.ReadBytes(4);
                        p.R = rgba[0];
                        p.G = rgba[1];
                        p.B = rgba[2];

                        p.Intensity = reader.ReadUInt16();

                        //Add direct to grid since we already established this to not be a leaf node
                        AddWithoutCheck(p);
                    }
                    catch
                    {
                        //We can assume if it fails to read one point the data is likely corrupted
                        break;
                    }
                }
            }

            reader.Close();
            reader = null;
            IsInMemory = true;
        }

        public void WriteToDisk(List<CloudPoint> points, bool append)
        {
            string filepath = WorkDir + "/data/" + "r/" + FilePath + ".bin";

            Directory.CreateDirectory(WorkDir + "/data/" + HierarchyPath);

            if (File.Exists(filepath))
                File.Delete(filepath);

            var byteArray = new byte[points.Count * 18];
            int offset = 0;
            var scale = 1 / Octree.Scale;
            var bounds = OffsetBounds;

            this.NumAccepted = (uint)points.Count;

            foreach (var p in points)
            {
                int flatX = (int)Math.Round(((p.X - bounds.Min.X) * scale));
                int flatY = (int)Math.Round(((p.Y - bounds.Min.Y) * scale));
                int flatZ = (int)Math.Round(((p.Z - bounds.Min.Z) * scale));

                //Write X
                byteArray[offset++] = (byte)flatX;
                byteArray[offset++] = (byte)(flatX >> 8);
                byteArray[offset++] = (byte)(flatX >> 16);
                byteArray[offset++] = (byte)(flatX >> 24);

                //Write Y
                byteArray[offset++] = (byte)flatY;
                byteArray[offset++] = (byte)(flatY >> 8);
                byteArray[offset++] = (byte)(flatY >> 16);
                byteArray[offset++] = (byte)(flatY >> 24);

                //Write Z
                byteArray[offset++] = (byte)flatZ;
                byteArray[offset++] = (byte)(flatZ >> 8);
                byteArray[offset++] = (byte)(flatZ >> 16);
                byteArray[offset++] = (byte)(flatZ >> 24);

                //Write RGBA
                byteArray[offset++] = p.R;
                byteArray[offset++] = p.G;
                byteArray[offset++] = p.B;
                byteArray[offset++] = 255;

                //Write Intensity
                byteArray[offset++] = (byte)(p.Intensity >> 8);
                byteArray[offset++] = (byte)p.Intensity;
            }

            //Convert points to byte array
            //Write the byte array instead (keeps creating the byte array of of the file access section)
            var writer = new BinaryWriter(File.Open(filepath, FileMode.Create));
            writer.Write(byteArray);

            writer.Close();
            writer = null;
        }

        public void Flush()
        {
            //Put write to disk logic inside of this method to improve performance? (Saves passing a list of numbers through and this is the only place write to disk is called
            try
            {
                if (IsLeafNode)
                {
                    if (AddCalledSinceLastFlush)
                        WriteToDisk(Store, false);
                    else if (IsInMemory)
                    {
                        Store.Clear();
                        Store = new List<CloudPoint>();
                        IsInMemory = false;
                    }
                }
                else
                {
                    if (AddCalledSinceLastFlush)
                    {
                        //Try not discarding points after adding to them? will cost resources to read the file over and over again
                        WriteToDisk(Grid.Select(p => p.Value).ToList(), true);
                        Grid.Clear();
                        Grid = new Dictionary<(int, int, int), CloudPoint>();
                        IsInMemory = false;
                    }
                    else if (IsInMemory)
                    {
                        Grid = new Dictionary<(int, int, int), CloudPoint>();
                        IsInMemory = false;
                    }
                }

                AddCalledSinceLastFlush = false;
                //if (Children != null)
                //{
                //    foreach (var child in Children)
                //    {
                //        if (child != null)
                //            child.Flush();
                //    }
                //}
            }
            catch
            {

            }
        }

        public List<CloudPoint> LoadIntoStore()
        { 
            var reader = new BinaryReader(File.Open(WorkDir + "/data/r/" + FilePath + ".bin", FileMode.Open, FileAccess.Read, FileShare.Read));
            var scale = Octree.Scale;
            var bounds = OffsetBounds;
            uint c = 0;
            uint totalPoints = (uint)(reader.BaseStream.Length / 18);

            var pointsList = new List<CloudPoint>((int)this.NumAccepted);

            while(c < totalPoints)
            {
                c++;
                try
                {
                    var p = new CloudPoint();
                    p.X = ((reader.ReadInt32() * scale) + bounds.Min.X);
                    p.Y = ((reader.ReadInt32() * scale) + bounds.Min.Y);
                    p.Z = ((reader.ReadInt32() * scale) + bounds.Min.Z);

                    //Alpha value is stored but not used so we must take next 4 bytes not 3
                    byte[] rgba = reader.ReadBytes(4);
                    p.R = rgba[0];
                    p.G = rgba[1];
                    p.B = rgba[2];

                    p.Intensity = reader.ReadUInt16();

                    pointsList.Add(p);
                }
                catch
                {
                    //We can assume if it fails to read one point the data is likely corrupted
                    break;
                }
            }

            reader.Close();
            reader = null;
            return pointsList;
        }

        public bool Add(CloudPoint point)
        {
            try
            {
                AddCalledSinceLastFlush = true;

                if (!IsInMemory)
                    LoadFromDisk();

                if (IsLeafNode)
                {
                    Store.Add(point);
                    Octree.TightBounds.Update(point.X, point.Y, point.Z);
                    NumAccepted++;

                    if (Store.Count >= StoreLimit)
                    {
                        NumAccepted = 0;
                        Split();
                    }

                    return true;
                }
                else
                {
                    //Check if dictionary with string is faster
                    var xI = (int)Math.Round((point.X - Bounds.Min.X) * SpacingFactor);
                    var yI = (int)Math.Round((point.Y - Bounds.Min.Y) * SpacingFactor);
                    var zI = (int)Math.Round((point.Z - Bounds.Min.Z) * SpacingFactor);

                    var key = (xI, yI, zI);
                    if(Grid.TryGetValue(key, out CloudPoint oldPoint))
                    {
                        //Check if the point is a duplicate or near enough
                        if (Math.Abs(oldPoint.X - point.X) <= 0.00001 && Math.Abs(oldPoint.Y - point.Y) <= 0.00001 && Math.Abs(oldPoint.Z - point.Z) <= 0.00001)
                            return false;

                        int childIndex = OctreeInterop.NodeIndex(Bounds, point);
                        if (childIndex >= 0)
                        {
                            if (IsLeafNode)
                                Children = new OctreeNode[8];

                            OctreeNode child = Children[childIndex];
                            if (child == null)
                                child = CreateChild(childIndex);

                            if (Grid.Keys.Count == 1)
                            {

                            }
                            return child.Add(point);
                        }
                        else
                            return false;
                    }
                    else
                    {
                        Grid.Add(key, point);
                        Octree.TightBounds.Update(point.X, point.Y, point.Z);
                        NumAccepted++;
                        return true;
                    }
                }
            }
            catch
            {
                return false;
            }
        }

        private void Split()
        {
            Children = new OctreeNode[8];
            string filePath = WorkDir + "/data/r/" + FilePath + ".bin";

            if (File.Exists(filePath))
            {
                //WA 13.02.19
                //Best guess is anti virus is locking up newly generated files meaning we can't delete them right away.
                //This loop just prevents the app from crashing while we wait to delete the file
                bool deleted = false;
                while (deleted == false)
                {
                    try
                    {
                        File.Delete(filePath);
                        deleted = true;
                    }
                    catch
                    {

                    }
                }
            }

            foreach (var p in Store)
            {
                Add(p);
            }

            Store.Clear();
            Store = new List<CloudPoint>();
        }

        /// <summary>
        /// Traverses all child nodes and sets added since last flush to false - legacy for file version 1
        /// </summary>
        /// <param name="startNode"></param>
        public void Traverse(OctreeNode startNode)
        {
            startNode.AddedSinceLastFlush = false;
            if (this.Children != null)
            {
                foreach (var child in Children)
                {
                    if (child != null)
                        child.Traverse(child);
                }
            }
        }

        public List<OctreeNode> GetHierarchy(int levels)
        {
            var hierarchy = new List<OctreeNode>();

            var stack = new List<OctreeNode>();

            stack.Add(this);
            while (stack.Count != 0)
            {
                OctreeNode node = stack.First();
                stack.RemoveAt(0);

                if (node.Level >= this.Level + levels)
                    break;

                hierarchy.Add(node);

                if (node.Children != null)
                {
                    foreach (var child in node.Children)
                    {
                        if (child != null)
                            stack.Add(child);
                    }
                }
            }

            return hierarchy;
        }

        public OctreeNode FindNode(string name)
        {
            string thisName = this.Name;

            if (name.Length == thisName.Length)
                return (name == thisName) ? this : null;
            else if (name.Length > thisName.Length)
            {
                //Convert sring to int for last character
                int childIndex = Convert.ToInt32(name.Substring(thisName.Length, 1));
                if (!IsLeafNode && Children[childIndex] != null)
                    return Children[childIndex].FindNode(name);
                else
                    return null;
            }
            else
                return null;
        }

        private OctreeNode CreateChild(int childIndex)
        {
            var cBounds = OctreeInterop.ChildBounds(Bounds, childIndex);
            OctreeNode child = new OctreeNode(Octree, childIndex, cBounds, Level + 1);
            child.Parent = this;
            Children[childIndex] = child;

            return child;
        }

        private void AddWithoutCheck(CloudPoint point)
        {
            try
            {
                var xI = (int)Math.Round((point.X - Bounds.Min.X) * SpacingFactor);
                var yI = (int)Math.Round((point.Y - Bounds.Min.Y) * SpacingFactor);
                var zI = (int)Math.Round((point.Z - Bounds.Min.Z) * SpacingFactor);

                Grid.Add((xI, yI, zI), point);
            }
            catch
            {
                //Duplicate point
            }
        }

        public Bounds OffsetBounds
        {
            get
            {
                if (Octree.Version == OctreeFileVersion.FileVersion1)
                    return this.Bounds;
                else
                    return Octree.OffsetBounds;
            }
        }

        private void SetupGrid()
        {
            GridSpacing = Spacing;
            GridHalfSpacing = GridSpacing / 2;
            GridSize = Math.Ceiling(Bounds.Size.MaxValue()).ToString().Length;
            this.Grid = new Dictionary<(int, int, int), CloudPoint>();
        }

        public void LoadVertexData(ColorType colorType, Bounds modelBounds, bool cancel, BackgroundWorker worker, Alignment alignment, ShapeControl shapeControl, System.Drawing.Color underBreakColor, System.Drawing.Color overBreakColor)
        {
            if (worker.CancellationPending)
                return;

            var fileName = Octree.WorkDir + "\\data\\r\\" + FilePath + ".bin";

            if(File.Exists(fileName))
            {
                //Load file in
                var path = fileName;
                var flatBounds = OffsetBounds;
                var scale = Octree.Scale;

                var reader = new BinaryReader(File.Open(path, FileMode.Open, FileAccess.Read, FileShare.Read));

                //Just do RGB if we can't do clearance
                if (alignment == null || shapeControl == null)
                {
                    LoadVertexData(ColorType.RGB, modelBounds, cancel, worker);
                    return;
                }

                //Read all bytes

                if (reader == null)
                    return;

                var bytes = reader.ReadBytes((int)reader.BaseStream.Length);
                reader.Close();

                //Now sort through bytes to get points

                var numberOfPoints = (uint)(bytes.Length / 18);
                uint numberOfReadPoints = 0;
                int byteIndex = 0, vertIndex = 0, colIndex = 0;
                var cX = (float)modelBounds.Center.X;
                var cY = (float)modelBounds.Center.Y;

                vertices = new float[numberOfPoints * 3];
                colors = new float[numberOfPoints * 3];
                float uR = underBreakColor.R / 255.0f, uG = underBreakColor.G / 255.0f, uB = underBreakColor.B / 255.0f;
                float oR = overBreakColor.R / 255.0f, oG = overBreakColor.G / 255.0f, oB = overBreakColor.B / 255.0f;

                while (numberOfReadPoints < numberOfPoints)
                {
                    var x = ((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.X);
                    vertices[vertIndex++] = (float)(x - cX);
                    byteIndex += 4;
                    var y = ((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Y);
                    vertices[vertIndex++] = (float)(y - cY);
                    byteIndex += 4;
                    var z = ((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Z);
                    vertices[vertIndex++] = (float)z;
                    byteIndex += 4;

                    //Calc the offset based on shape

                    double dist = 0, offset = 0;
                    double closestOffset = double.MaxValue;

                    var outVec = new AlignmentVector5();
                    if (alignment.Horizontal.GetChainageAndOffset(new Point2D(x, y), ref outVec))
                    {
                        var grade = alignment.Vertical.GradeLevel(outVec.chainage, out double vAngle);
                        var shapeWithCant = shapeControl.GetShape(outVec.chainage, alignment, grade, alignment.GetCantAtChainage(outVec.chainage), vAngle, ShapeType.Design);

                        if (shapeWithCant != null)
                        {
                            //Find closest element to alignment
                            foreach (var element in shapeWithCant.Elements)
                            {
                                var elementDistance = element.Length;

                                if (element.Radius != 0)
                                    Trig.DistanceAndOffsetFromLine(element.StartX, element.StartY, element.EndX, element.EndY, element.Radius, outVec.offset, z, ref dist, ref offset);
                                else
                                {
                                    double tBrg = 0, tDist = 0;
                                    Trig.RPC(element.StartX, element.StartY, element.EndX, element.EndY, ref tBrg, ref tDist);
                                    Trig.DistanceAndOffsetFromLine(element.StartX, element.StartY, tBrg, outVec.offset, z, ref dist, ref offset);
                                }

                                if (dist >= 0 && dist <= elementDistance)
                                {
                                    if (Math.Abs(offset) < Math.Abs(closestOffset))
                                        closestOffset = offset;
                                }
                            }

                            //Second pass to check distance from tangent points
                            foreach (var element in shapeWithCant.Elements)
                            {
                                var distToTangent = Vector.FnDistance(element.StartX, element.StartY, outVec.offset, z);
                                if (distToTangent < Math.Abs(closestOffset))
                                {
                                    if (outVec.offset < 0)
                                        closestOffset = outVec.offset <= element.StartX ? distToTangent * -1 : distToTangent;
                                    else
                                        closestOffset = outVec.offset >= element.StartX ? distToTangent * -1 : distToTangent;
                                }
                            }

                            //Now we have the closest 
                            if (closestOffset < 0 || closestOffset == double.MaxValue)
                            {
                                colors[colIndex++] = oR;
                                colors[colIndex++] = oG;
                                colors[colIndex++] = oB;
                            }
                            else
                            {
                                colors[colIndex++] = uR;
                                colors[colIndex++] = uG;
                                colors[colIndex++] = uB;
                            }
                        }
                        else
                        {
                            //Set color to black
                            colors[colIndex++] = 0.0f;
                            colors[colIndex++] = 0.0f;
                            colors[colIndex++] = 0.0f;
                        }
                    }
                    else
                    {
                        //Set color to black
                        colors[colIndex++] = 0.0f;
                        colors[colIndex++] = 0.0f;
                        colors[colIndex++] = 0.0f;
                    }

                    byteIndex += 6;
                    numberOfReadPoints++;
                }
            }
        }

        public void LoadVertexData(ColorType colorType, Bounds modelBounds, bool cancel, BackgroundWorker worker)
        {
            if (worker.CancellationPending)
                return;

            var fileName = Octree.WorkDir + "\\data\\r\\" + FilePath + ".bin";

            if(File.Exists(fileName))
            {
                //Load file in
                var path = fileName;
                var flatBounds = OffsetBounds;
                var scale = Octree.Scale;

                var reader = new BinaryReader(File.Open(path, FileMode.Open, FileAccess.Read, FileShare.Read));

                //Read all bytes
                if (reader == null)
                    return;

                var bytes = reader.ReadBytes((int)reader.BaseStream.Length);
                reader.Close();

                //Intensity stuff
                ushort intensity = 0;
                int minIntensity = Octree.IntensityConverter.MinIntensity, maxIntensity = Octree.IntensityConverter.MaxIntensity;
                int outOfRangeMode = Octree.IntensityConverter.OutOfRangeMode;
                var firstColor = Octree.IntensityConverter.IntensityColorList.First().StartColor;
                var lastColor = Octree.IntensityConverter.IntensityColorList.Last().EndColor;
                var cX = (float)(modelBounds.Center.X);
                var cY = (float)(modelBounds.Center.Y);

                //Now sort through bytes to get the points
                var numberOfPoints = (uint)(bytes.Length / 18);
                uint numberOfReadPoints = 0;
                int byteIndex = 0, vertIndex = 0, colIndex = 0;

                vertices = new float[numberOfPoints * 3];
                colors = new float[numberOfPoints * 3];
                float colScale = 1f / 255f;
                int cR = 0, cG = 0, cB = 0;

                switch (colorType)
                {
                    case ColorType.RGB:
                        while (numberOfReadPoints < numberOfPoints)
                        {
                            vertices[vertIndex++] = (float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.X) - cX);
                            byteIndex += 4;
                            vertices[vertIndex++] = (float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Y) - cY);
                            byteIndex += 4;
                            vertices[vertIndex++] = (float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Z));
                            byteIndex += 4;

                            colors[colIndex++] = bytes[byteIndex++] * colScale;
                            colors[colIndex++] = bytes[byteIndex++] * colScale;
                            colors[colIndex++] = bytes[byteIndex++] * colScale;

                            byteIndex += 3;
                            numberOfReadPoints++;
                        }
                        return;
                    case ColorType.HeightMap:
                        while (numberOfReadPoints < numberOfPoints)
                        {
                            vertices[vertIndex++] = (float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.X) - cX);
                            byteIndex += 4;
                            vertices[vertIndex++] = (float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Y) - cY);
                            byteIndex += 4;
                            var z = (float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Z));
                            vertices[vertIndex++] = (float)z;
                            byteIndex += 4;

                            Services.Conversions.GetHeightMapFromPoint(modelBounds, z, ref cR, ref cG, ref cB);

                            colors[colIndex++] = cR * colScale;
                            colors[colIndex++] = cG * colScale;
                            colors[colIndex++] = cB * colScale;

                            byteIndex += 6;
                            numberOfReadPoints++;
                        }
                        return;
                    case ColorType.GrayScale:
                        while (numberOfReadPoints < numberOfPoints)
                        {
                            vertices[vertIndex++] = (float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.X) - cX);
                            byteIndex += 4;
                            vertices[vertIndex++] = (float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Y) - cY);
                            byteIndex += 4;
                            vertices[vertIndex++] = (float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Z));
                            byteIndex += 4;

                            var gray = Services.Conversions.GetGrayScaleFromColor(bytes[byteIndex++], bytes[byteIndex++], bytes[byteIndex++]) * colScale;

                            colors[colIndex++] = gray;
                            colors[colIndex++] = gray;
                            colors[colIndex++] = gray;

                            byteIndex += 3;
                            numberOfReadPoints++;
                        }
                        return;
                    case ColorType.Intensity:
                        var verts = new List<float>((int)numberOfPoints * 3);
                        var cols = new List<float>((int)numberOfPoints * 3);

                        while (numberOfReadPoints < numberOfPoints)
                        {
                            //Get intensity and check that it is within range
                            intensity = BitConverter.ToUInt16(bytes, byteIndex + 16);

                            if(intensity < minIntensity || intensity > maxIntensity)
                            {
                                //Intensity is out of range
                                if(outOfRangeMode == 0)
                                {
                                    //Show RGB
                                    verts.Add((float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.X) - cX));
                                    byteIndex += 4;
                                    verts.Add((float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Y) - cY));
                                    byteIndex += 4;
                                    verts.Add((float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Z)));
                                    byteIndex += 4;

                                    cols.Add(bytes[byteIndex++] * colScale);
                                    cols.Add(bytes[byteIndex++] * colScale);
                                    cols.Add(bytes[byteIndex++] * colScale);

                                    byteIndex += 3;
                                    numberOfReadPoints++;
                                    continue;
                                }
                                else if(outOfRangeMode == 1)
                                {
                                    //Min / Max intensity
                                    verts.Add((float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.X) - cX));
                                    byteIndex += 4;
                                    verts.Add((float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Y) - cY));
                                    byteIndex += 4;
                                    verts.Add((float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Z)));
                                    byteIndex += 10;

                                    if(intensity < minIntensity)
                                    {
                                        cols.Add(firstColor.R * colScale);
                                        cols.Add(firstColor.G * colScale);
                                        cols.Add(firstColor.B * colScale);
                                    }
                                    else if(intensity > maxIntensity)
                                    {
                                        cols.Add(lastColor.R * colScale);
                                        cols.Add(lastColor.G * colScale);
                                        cols.Add(lastColor.B * colScale);
                                    }

                                    numberOfReadPoints++;
                                    continue;
                                }
                                else if(outOfRangeMode == 2)
                                {
                                    //Hide the point so basically don't draw it
                                    byteIndex += 18;
                                    numberOfReadPoints++;
                                    continue;
                                }
                            }
                            else
                            {
                                Octree.IntensityConverter.GetIntensityColor(intensity, ref cR, ref cG, ref cB);

                                verts.Add((float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.X) - cX));
                                byteIndex += 4;
                                verts.Add((float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Y) - cY));
                                byteIndex += 4;
                                verts.Add((float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Z)));
                                byteIndex += 10;

                                cols.Add(cR * colScale);
                                cols.Add(cG * colScale);
                                cols.Add(cB * colScale);

                                numberOfReadPoints++;
                                continue;
                            }
                        }

                        vertices = verts.ToArray();
                        verts.Clear();

                        colors = cols.ToArray();
                        cols.Clear();

                        return;
                }
            }
        }

        public void LoadVertexData(ColorType colorType, Bounds modelBounds, bool cancel, BackgroundWorker worker, List<Polygon> selectionZones, List<Polygon> deselectionZones)
        {
            if (worker.CancellationPending)
                return;

            var polysToCheck = new List<Polygon>();
            var deselectPolysToCheck = new List<Polygon>();

            foreach(var poly in selectionZones)
            {
                if (Bounds.Intersects2D(poly.SquareBounds) && !poly.Filtered)
                    polysToCheck.Add(poly);
            }

            if (polysToCheck.Count <= 0)
                return;

            foreach(var poly in deselectionZones)
            {
                if (Bounds.Intersects2D(poly.SquareBounds))
                    deselectPolysToCheck.Add(poly);
            }

            var fileName = Octree.WorkDir + "\\data\\r\\" + FilePath + ".bin";

            if(File.Exists(fileName))
            {
                //Load file in
                var path = fileName;
                var flatBounds = OffsetBounds;
                var scale = Octree.Scale;

                var reader = new BinaryReader(File.Open(path, FileMode.Open, FileAccess.Read, FileShare.Read));

                //Read all bytes
                if (reader == null)
                {
                    vertices = null;
                    colors = null;
                    return;
                }

                var bytes = reader.ReadBytes((int)reader.BaseStream.Length);
                reader.Close();

                //Now sort through bytes to get the points
                var numberOfPoints = (uint)(bytes.Length / 18);
                uint numberOfReadPoints = 0;
                int byteIndex = 0;
                bool load = true;

                var verts = new List<float>((int)numberOfPoints);
                var cols = new List<float>((int)numberOfPoints);
                var cX = (float)modelBounds.Center.X;
                var cY = (float)modelBounds.Center.Y;

                //Intensity stuff
                ushort intensity = 0;
                int minIntensity = Octree.IntensityConverter.MinIntensity, maxIntensity = Octree.IntensityConverter.MaxIntensity;
                int outOfRangeMode = Octree.IntensityConverter.OutOfRangeMode;
                var firstColor = Octree.IntensityConverter.IntensityColorList.First().StartColor;
                var lastColor = Octree.IntensityConverter.IntensityColorList.Last().EndColor;

                float colScale = 1f / 255f;
                int cR = 0, cG = 0, cB = 0;

                double x = 0, y = 0, z = 0;
                switch (colorType)
                {
                    case ColorType.RGB:
                        while (numberOfReadPoints < numberOfPoints)
                        {
                            load = false;

                            x = ((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.X);
                            byteIndex += 4;
                            y = ((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Y);
                            byteIndex += 4;

                            foreach (var poly in selectionZones)
                            {
                                if (poly.InPoly2D(x, y, 0))
                                {
                                    load = true;
                                    foreach (var deselectPoly in deselectionZones)
                                    {
                                        if (deselectPoly.InPoly2D(x, y, 0))
                                        {
                                            load = false;
                                            break;
                                        }
                                    }

                                    break;
                                }
                            }

                            numberOfReadPoints++;

                            if (load)
                            {
                                verts.Add((float)(x - cX));
                                verts.Add((float)(y - cY));
                                verts.Add((float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Z)));
                                byteIndex += 4;

                                cols.Add(bytes[byteIndex++] * colScale);
                                cols.Add(bytes[byteIndex++] * colScale);
                                cols.Add(bytes[byteIndex++] * colScale);

                                byteIndex += 3;
                            }
                            else
                                byteIndex += 10;
                        }
                        break;
                    case ColorType.HeightMap:
                        while (numberOfReadPoints < numberOfPoints)
                        {
                            load = false;

                            x = ((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.X);
                            byteIndex += 4;
                            y = ((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Y);
                            byteIndex += 4;

                            foreach (var poly in selectionZones)
                            {
                                if (poly.InPoly2D(x, y, 0))
                                {
                                    load = true;
                                    foreach (var deselectPoly in deselectionZones)
                                    {
                                        if (deselectPoly.InPoly2D(x, y, 0))
                                        {
                                            load = false;
                                            break;
                                        }
                                    }

                                    break;
                                }
                            }

                            numberOfReadPoints++;

                            if (load)
                            {
                                z = ((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Z);
                                byteIndex += 4;

                                verts.Add((float)(x - cX));
                                verts.Add((float)(y - cY));
                                verts.Add((float)z);

                                Conversions.GetHeightMapFromPoint(modelBounds, z, ref cR, ref cG, ref cB);

                                cols.Add(cR * colScale);
                                cols.Add(cG * colScale);
                                cols.Add(cB * colScale);

                                byteIndex += 6;
                            }
                            else
                                byteIndex += 10;
                        }
                        break;
                    case ColorType.GrayScale:
                        while (numberOfReadPoints < numberOfPoints)
                        {
                            load = false;

                            x = ((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.X);
                            byteIndex += 4;
                            y = ((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Y);
                            byteIndex += 4;

                            foreach (var poly in selectionZones)
                            {
                                if (poly.InPoly2D(x, y, 0))
                                {
                                    load = true;
                                    foreach (var deselectPoly in deselectionZones)
                                    {
                                        if (deselectPoly.InPoly2D(x, y, 0))
                                        {
                                            load = false;
                                            break;
                                        }
                                    }

                                    break;
                                }
                            }

                            numberOfReadPoints++;

                            if (load)
                            {
                                verts.Add((float)(x - cX));
                                verts.Add((float)(y - cY));
                                verts.Add((float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Z)));
                                byteIndex += 4;

                                var gray = Services.Conversions.GetGrayScaleFromColor(bytes[byteIndex++], bytes[byteIndex++], bytes[byteIndex++]) * colScale;

                                cols.Add(gray);
                                cols.Add(gray);
                                cols.Add(gray);
                                byteIndex += 3;
                            }
                            else
                                byteIndex += 10;
                        }
                        break;
                    case ColorType.Intensity:
                        while (numberOfReadPoints < numberOfPoints)
                        {
                            load = false;

                            x = ((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.X);
                            byteIndex += 4;
                            y = ((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Y);
                            byteIndex += 4;

                            foreach (var poly in selectionZones)
                            {
                                if (poly.InPoly2D(x, y, 0))
                                {
                                    load = true;
                                    foreach (var deselectPoly in deselectionZones)
                                    {
                                        if (deselectPoly.InPoly2D(x, y, 0))
                                        {
                                            load = false;
                                            break;
                                        }
                                    }

                                    break;
                                }
                            }

                            numberOfReadPoints++;

                            if (load)
                            {
                                //X, Y already loaded at this point
                                intensity = BitConverter.ToUInt16(bytes, byteIndex + 8);
                                
                                if(intensity < minIntensity || intensity > maxIntensity)
                                {
                                    //Intensity is out of range
                                    if(outOfRangeMode == 0)
                                    {
                                        //Show RGB
                                        verts.Add((float)(x - cX));
                                        verts.Add((float)(y - cY));
                                        verts.Add((float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Z)));
                                        byteIndex += 4;

                                        cols.Add(bytes[byteIndex++] * colScale);
                                        cols.Add(bytes[byteIndex++] * colScale);
                                        cols.Add(bytes[byteIndex++] * colScale);

                                        byteIndex += 3;
                                        continue;
                                    }
                                    else if(outOfRangeMode == 1)
                                    {
                                        //Min / Max Intensity
                                        verts.Add((float)(x - cX));
                                        verts.Add((float)(y - cY));
                                        verts.Add((float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Z)));
                                        byteIndex += 10;

                                        if(intensity < minIntensity)
                                        {
                                            cols.Add(firstColor.R * colScale);
                                            cols.Add(firstColor.G * colScale);
                                            cols.Add(firstColor.B * colScale);
                                        }
                                        else if(intensity > maxIntensity)
                                        {
                                            cols.Add(lastColor.R * colScale);
                                            cols.Add(lastColor.G * colScale);
                                            cols.Add(lastColor.B * colScale);
                                        }

                                        continue;
                                    }
                                    else if(outOfRangeMode == 2)
                                    {
                                        //Hide the point
                                        byteIndex += 10;
                                        continue;
                                    }
                                }
                                else
                                {
                                    //Point is within the intensity range so draw it normally
                                    Octree.IntensityConverter.GetIntensityColor(intensity, ref cR, ref cG, ref cB);

                                    verts.Add((float)(x - cX));
                                    verts.Add((float)(y - cY));
                                    verts.Add((float)(((BitConverter.ToInt32(bytes, byteIndex) * scale) + flatBounds.Min.Z)));
                                    byteIndex += 10;

                                    cols.Add(cR * colScale);
                                    cols.Add(cG * colScale);
                                    cols.Add(cB * colScale);
                                    continue;
                                }
                            }
                            else
                                byteIndex += 10;
                        }
                        break;
                }

                //If there is data convert to float arrays and return
                if (verts.Count >= 0 && cols.Count >= 0)
                {
                    vertices = verts.ToArray();
                    colors = cols.ToArray();
                }
                else
                {
                    vertices = null;
                    colors = null;
                    verts.Clear();
                    cols.Clear();
                    return;
                }
            }
        }

        public void  GetAllNodes(ref List<OctreeNode> nodes)
        {
            nodes.Add(this);

            if(Children != null)
            {
                foreach(var child in Children)
                {
                    if (child != null)
                        child.GetAllNodes(ref nodes);
                }
            }
        }

        public void GetAllNodesToFlush(ref List<OctreeNode> nodes)
        {
            if (this.IsInMemory)
                nodes.Add(this);

            if(Children != null)
            {
                foreach(var child in Children)
                {
                    if (child != null)
                        child.GetAllNodesToFlush(ref nodes);
                }
            }
        }

        #endregion

        #region Data Extraction

        public bool GetClosest(Ray ray, ref double x, ref double y, ref double z, ref byte r, ref byte g, ref byte b, ref ushort intensity, ref double distance)
        {
            if (Bounds.IntersectsWithRay(ray))
            {
                var points = LoadIntoStore();
                foreach (var p in points)
                {
                    var newDist = Vector.ClosestDistanceOnALine(ray.Origin.X, ray.Origin.Y, ray.Origin.Z, ray.End.X, ray.End.Y, ray.End.Z, p.X, p.Y, p.Z);
                    if (newDist < distance)
                    {
                        x = p.X;
                        y = p.Y;
                        z = p.Z;
                        r = p.R;
                        g = p.G;
                        b = p.B;
                        intensity = p.Intensity;
                        distance = newDist;
                    }
                }
                points.Clear();
                //points = new List<Point>();
                //grid = null;


                if (Children != null)
                {
                    foreach (var child in Children)
                    {
                        if (child != null)
                        {
                            child.GetClosest(ray, ref x, ref y, ref z, ref r, ref g, ref b, ref intensity, ref distance);
                        }
                    }
                }
                return true;
            }
            else
            {
                return false;
            }
        }

        public bool GetClosest(Ray ray, ref double x, ref double y, ref double z, ref byte r, ref byte g, ref byte b, ref ushort intensity, ref double distance, ref BackgroundWorker worker, DoWorkEventArgs e)
        {
            if (worker.CancellationPending)
                return false;

            if (Bounds.IntersectsWithRay(ray))
            {
                if (worker.CancellationPending)
                    return false;

                var points = LoadIntoStore();
                if (worker.CancellationPending)
                {
                    return false;
                }
                foreach (var p in points)
                {
                    var newDist = Vector.ClosestDistanceOnALine(ray.Origin.X, ray.Origin.Y, ray.Origin.Z, ray.End.X, ray.End.Y, ray.End.Z, p.X, p.Y, p.Z);
                    if (newDist < distance)
                    {
                        x = p.X;
                        y = p.Y;
                        z = p.Z;
                        r = p.R;
                        g = p.G;
                        b = p.B;
                        intensity = p.Intensity;
                        distance = newDist;
                    }
                }
                points.Clear();

                if (worker.CancellationPending)
                    return false;

                if (Children != null)
                {
                    foreach (var child in Children)
                    {
                        if (child != null)
                            child.GetClosest(ray, ref x, ref y, ref z, ref r, ref g, ref b, ref intensity, ref distance, ref worker, e);
                    }
                }
                return true;
            }
            else
                return false;
        }

        public bool GetClosestPointFromSection(Ray ray, Bounds sectionBounds, ref double x, ref double y, ref double z, ref byte r, ref byte g, ref byte b, ref ushort intensity, ref double distance)
        {
            if (Bounds.IntersectsWithRay(ray) && Bounds.Intersects2D(sectionBounds))
            {
                var points = LoadIntoStore();
                foreach (var p in points)
                {
                    if (sectionBounds.IsInside(p.X, p.Y, p.Z))
                    {
                        var newDist = Vector.ClosestDistanceOnALine(ray.Origin.X, ray.Origin.X, ray.Origin.Z, ray.End.X, ray.End.Y, ray.End.Z, p.X, p.Y, p.Z);
                        if (newDist < distance)
                        {
                            x = p.X;
                            y = p.Y;
                            z = p.Z;
                            r = p.R;
                            g = p.G;
                            b = p.B;
                            intensity = p.Intensity;
                            distance = newDist;
                        }
                    }
                }
                points.Clear();
                if (Children != null)
                {
                    foreach (var child in Children)
                    {
                        if (child != null)
                        {
                            child.GetClosestPointFromSection(ray, sectionBounds, ref x, ref y, ref z, ref r, ref g, ref b, ref intensity, ref distance);
                        }
                    }
                }
                return true;
            }
            return false;
        }

        public bool GetClosestPointFromSection(Ray ray, Polygon sectionPoly, ref double x, ref double y, ref double z, ref byte r, ref byte g, ref byte b, ref ushort intensity, ref double distance)
        {
            if (Bounds.IntersectsWithRay(ray) && Bounds.Intersects2D(sectionPoly.SquareBounds))
            {
                var points = LoadIntoStore();
                foreach (var p in points)
                {
                    if (sectionPoly.InPoly2D(p.X, p.Y, p.Z))
                    {
                        var newDist = Vector.ClosestDistanceOnALine(ray.Origin.X, ray.Origin.Y, ray.Origin.Z, ray.End.X, ray.End.Y, ray.End.Z, p.X, p.Y, p.Z);
                        if (newDist < distance)
                        {
                            x = p.X;
                            y = p.Y;
                            z = p.Z;
                            r = p.R;
                            g = p.G;
                            b = p.B;
                            intensity = p.Intensity;
                            distance = newDist;
                        }
                    }
                }
                points.Clear();

                if (Children != null)
                {
                    foreach (var child in Children)
                    {
                        if (child != null)
                        {
                            child.GetClosestPointFromSection(ray, sectionPoly, ref x, ref y, ref z, ref r, ref g, ref b, ref intensity, ref distance);
                        }
                    }
                }
                return true;
            }
            return false;
        }

        public bool GetClosestPointFromSelectionSection(Ray ray, Polygon sectionPoly, List<Polygon> selectionZones, List<Polygon> deselectionZones, ref double x, ref double y, ref double z, ref byte r, ref byte g, ref byte b, ref ushort intensity, ref double distance)
        {
            if (Bounds.IntersectsWithRay(ray) && Bounds.Intersects2D(sectionPoly.SquareBounds))
            {
                bool cont = false;
                //Check if bounds intersects with any selectionzones
                List<Polygon> polysToCheck = new List<Polygon>();
                List<Polygon> deselectPolysToCheck = new List<Polygon>();
                foreach (var zone in selectionZones)
                {
                    if (Bounds.Intersects2D(zone.SquareBounds))
                        polysToCheck.Add(zone);
                }

                if (polysToCheck.Count <= 0)
                    return false;

                foreach (var zone in deselectionZones)
                {
                    if (Bounds.Intersects2D(zone.SquareBounds))
                        deselectPolysToCheck.Add(zone);
                }

                var points = LoadIntoStore();
                foreach (var p in points)
                {
                    if (sectionPoly.InPoly2D(p.X, p.Y, p.Z))
                    {
                        cont = false;
                        //Check its selected
                        foreach (var zone in polysToCheck)
                        {
                            if (zone.InPoly2D(p.X, p.Y, p.Z))
                            {
                                cont = true;
                                break;
                            }
                        }

                        if (!cont)
                            continue;

                        //Check its not deselected
                        cont = true;
                        foreach (var zone in deselectPolysToCheck)
                        {
                            if (zone.InPoly2D(p.X, p.Y, p.Z))
                            {
                                cont = false;
                                break;
                            }
                        }

                        if (!cont)
                            continue;

                        //If we made it to here its both selected point and not deselected get distance and check value
                        var newDist = Vector.ClosestDistanceOnALine(ray.Origin.X, ray.Origin.Y, ray.Origin.Z, ray.End.X, ray.End.Y, ray.End.Z, p.X, p.Y, p.Z);
                        if (newDist < distance)
                        {
                            x = p.X;
                            y = p.Y;
                            z = p.Z;
                            r = p.R;
                            g = p.G;
                            b = p.B;
                            intensity = p.Intensity;
                            distance = newDist;
                        }
                    }
                }
                points.Clear();

                if (Children != null)
                {
                    foreach (var child in Children)
                    {
                        if (child != null)
                        {
                            child.GetClosestPointFromSelectionSection(ray, sectionPoly, selectionZones, deselectionZones, ref x, ref y, ref z, ref r, ref g, ref b, ref intensity, ref distance);
                        }
                    }
                }
                return true;
            }
            else
                return false;

        }

        public bool GetClosestPointFromSelection(Ray ray, List<Polygon> selectionZones, List<Polygon> deselectionZones, ref double x, ref double y, ref double z, ref byte r, ref byte g, ref byte b, ref ushort intensity, ref double distance)
        {
            if (Bounds.IntersectsWithRay(ray))
            {
                bool cont = false;
                //Check if bounds intersects with any selectionzones
                List<Polygon> polysToCheck = new List<Polygon>();
                List<Polygon> deselectPolysToCheck = new List<Polygon>();
                foreach (var zone in selectionZones)
                {
                    if (Bounds.Intersects2D(zone.SquareBounds))
                        polysToCheck.Add(zone);
                }

                if (polysToCheck.Count <= 0)
                    return false;

                foreach (var zone in deselectionZones)
                {
                    if (Bounds.Intersects2D(zone.SquareBounds))
                        deselectPolysToCheck.Add(zone);
                }


                var points = LoadIntoStore();
                foreach (var p in points)
                {
                    cont = false;
                    //Check its selected
                    foreach (var zone in polysToCheck)
                    {
                        if (zone.InPoly2D(p.X, p.Y, p.Z))
                        {
                            cont = true;
                            break;
                        }
                    }

                    if (!cont)
                        continue;

                    //Check its not deselected
                    cont = true;
                    foreach (var zone in deselectPolysToCheck)
                    {
                        if (zone.InPoly2D(p.X, p.Y, p.Z))
                        {
                            cont = false;
                            break;
                        }
                    }

                    if (!cont)
                        continue;

                    //If we made it to here its both selected point and no deselected, get distance and check value
                    var newDist = Vector.ClosestDistanceOnALine(ray.Origin.X, ray.Origin.Y, ray.Origin.Z, ray.End.X, ray.End.Y, ray.End.Z, p.X, p.Y, p.Z);
                    if (newDist < distance)
                    {
                        x = p.X;
                        y = p.Y;
                        z = p.Z;
                        r = p.R;
                        g = p.G;
                        b = p.B;
                        intensity = p.Intensity;
                        distance = newDist;
                    }
                }
                points.Clear();

                if (Children != null)
                {
                    foreach (var child in Children)
                    {
                        if (child != null)
                        {
                            child.GetClosestPointFromSelection(ray, selectionZones, deselectionZones, ref x, ref y, ref z, ref r, ref g, ref b, ref intensity, ref distance);
                        }
                    }
                }
                return true;
            }
            else
            {
                return false;
            }
        }

        public bool GetClosestPointFromSelection(Ray ray, List<Polygon> selectionZones, List<Polygon> deselectionZones, ref double x, ref double y, ref double z, ref byte r, ref byte g, ref byte b, ref ushort intensity, ref double distance, ref BackgroundWorker worker, DoWorkEventArgs e)
        {
            if (Bounds.IntersectsWithRay(ray))
            {
                if (worker.CancellationPending)
                {
                    return false;
                }

                bool cont = false;
                //Check if bounds intersections with any selectionZones

                List<Polygon> polysToCheck = new List<Polygon>();
                List<Polygon> deselectPolysToCheck = new List<Polygon>();
                foreach (var zone in selectionZones)
                {
                    if (Bounds.Intersects2D(zone.SquareBounds))
                        polysToCheck.Add(zone);
                }

                if (polysToCheck.Count <= 0)
                    return false;

                foreach (var zone in deselectionZones)
                {
                    if (Bounds.Intersects2D(zone.SquareBounds))
                        deselectPolysToCheck.Add(zone);
                }

                var points = LoadIntoStore();
                foreach (var p in points)
                {
                    cont = false;
                    //Check its selected
                    foreach (var zone in polysToCheck)
                    {
                        if (zone.InPoly2D(p.X, p.Y, p.Z))
                        {
                            cont = true;
                            break;
                        }
                    }

                    if (!cont)
                        continue;

                    //Check its not deselected
                    foreach (var zone in deselectPolysToCheck)
                    {
                        if (zone.InPoly2D(p.X, p.Y, p.Z))
                        {
                            cont = false;
                            break;
                        }
                    }

                    if (!cont)
                        continue;

                    //If we made it to here its both selected and not deselected, get distance and check value
                    var newDist = Vector.ClosestDistanceOnALine(ray.Origin.X, ray.Origin.Y, ray.Origin.Z, ray.End.X, ray.End.Y, ray.End.Z, p.X, p.Y, p.Z);
                    if (newDist < distance)
                    {
                        x = p.X;
                        y = p.Y;
                        z = p.Y;
                        r = p.R;
                        g = p.G;
                        b = p.B;
                        intensity = p.Intensity;
                        distance = newDist;
                    }
                }
                points.Clear();

                if (Children != null)
                {
                    foreach (var child in Children)
                        if (child != null)
                            child.GetClosestPointFromSelection(ray, selectionZones, deselectionZones, ref x, ref y, ref z, ref r, ref g, ref b, ref intensity, ref distance, ref worker, e);
                }
                return true;
            }
            else
                return false;
        }

        public bool GetSectionFromSelectionWithClosest(Ray ray, ref CrossSection section, List<Polygon> selectionZones, List<Polygon> deselectionZones, ref double distance, ref BackgroundWorker worker, DoWorkEventArgs e)
        {
            //Get all points within section bounds
            if (Bounds.Intersects2D(section.Poly.SquareBounds))
            {
                if (worker.CancellationPending == true)
                {
                    e.Cancel = true;
                    return false;
                }

                List<Polygon> polysToCheck = new List<Polygon>();
                List<Polygon> deselectPolysToCheck = new List<Polygon>();

                foreach (var zone in selectionZones)
                {
                    if (Bounds.Intersects2D(zone.SquareBounds))
                        polysToCheck.Add(zone);
                }

                if (polysToCheck.Count <= 0)
                    return false;

                foreach (var zone in deselectionZones)
                {
                    if (Bounds.Intersects2D(zone.SquareBounds))
                        deselectPolysToCheck.Add(zone);
                }

                //Get all points within the section
                var points = LoadIntoStore();
                bool cont;

                if (Bounds.IntersectsWithRay(ray))
                {
                    foreach (var p in points)
                    {
                        if (section.Poly.InPoly2D(p.X, p.Y, p.Z))
                        {
                            cont = false;
                            //Check its selected
                            foreach (var zone in polysToCheck)
                            {
                                if (zone.InPoly2D(p.X, p.Y, p.Z))
                                {
                                    cont = true;
                                    break;
                                }
                            }

                            if (!cont)
                                continue;

                            //Check its not deselected
                            cont = true;
                            foreach (var zone in deselectPolysToCheck)
                            {
                                if (zone.InPoly2D(p.X, p.Y, p.Z))
                                {
                                    cont = false;
                                    break;
                                }
                            }

                            if (!cont)
                                continue;

                            //Point is selected and not deselected
                            var newDist = Vector.ClosestDistanceOnALine(ray.Origin.X, ray.Origin.Y, ray.Origin.Z, ray.End.X, ray.End.Y, ray.End.Z, p.X, p.Y, p.Z);
                            if (newDist < distance)
                            {
                                section.ClosestPoint = new CloudPoint(p.X, p.Y, p.Z, p.R, p.G, p.B);
                                distance = newDist;
                            }

                            double newX = 0, newY = 0;
                            Trig.DistanceAndOffsetFromLine(section.SectionCenter.X, section.SectionCenter.Y, section.Bearing, p.X, p.Y, ref newX, ref newY);
                            section.SectionBounds.Update(newX, p.Z, newY);
                            section.OriginalPoints.Add(new SectionPoint(newX, p.Z, newY, p.R, p.G, p.B));
                        }
                    }
                }
                else
                {
                    foreach (var p in points)
                    {
                        if (section.Poly.InPoly2D(p.X, p.Y, p.Z))
                        {
                            cont = false;
                            //Check its selected
                            foreach (var zone in polysToCheck)
                            {
                                if (zone.InPoly2D(p.X, p.Y, p.Z))
                                {
                                    cont = true;
                                    break;
                                }
                            }

                            if (!cont)
                                continue;

                            //Check its not deselected
                            cont = true;
                            foreach (var zone in deselectPolysToCheck)
                            {
                                if (zone.InPoly2D(p.X, p.Y, p.Z))
                                {
                                    cont = false;
                                    break;
                                }
                            }

                            if (!cont)
                                continue;

                            //Point is selected and not deselected
                            double newX = 0, newY = 0;
                            Trig.DistanceAndOffsetFromLine(section.SectionCenter.X, section.SectionCenter.Y, section.Bearing, p.X, p.Y, ref newX, ref newY);
                            section.OriginalPoints.Add(new SectionPoint(newX, p.Z, newY, p.R, p.G, p.B));
                            section.SectionBounds.Update(newX, p.Z, newY);
                        }
                    }
                }

                points.Clear();

                if (Children != null)
                {
                    foreach (var child in Children)
                    {
                        if (child != null)
                            child.GetSectionFromSelectionWithClosest(ray, ref section, selectionZones, deselectionZones, ref distance, ref worker, e);
                    }
                }

                return true;
            }
            else
                return false;
        }

        public bool GetSectionFromSelectionWithoutClosest(ref CrossSection section, List<Polygon> selectionZones, List<Polygon> deselectionZones, ref BackgroundWorker worker, DoWorkEventArgs e)
        {
            //Get all points within sectionBounds
            if (Bounds.Intersects2D(section.Poly.SquareBounds))
            {
                if (worker.CancellationPending == true)
                {
                    e.Cancel = true;
                    return false;
                }

                List<Polygon> polysToCheck = new List<Polygon>();
                List<Polygon> deselectPolysToCheck = new List<Polygon>();

                foreach (var zone in selectionZones)
                {
                    if (Bounds.Intersects2D(zone.SquareBounds))
                        polysToCheck.Add(zone);
                }

                foreach (var zone in deselectionZones)
                {
                    if (Bounds.Intersects2D(zone.SquareBounds))
                        deselectPolysToCheck.Add(zone);
                }

                //Get all points within the section
                var points = LoadIntoStore();
                bool cont;

                foreach (var p in points)
                {
                    //If its within the section
                    if (section.Poly.InPoly2D(p.X, p.Y, p.Z))
                    {
                        cont = false;

                        //Check its selected
                        foreach (var zone in polysToCheck)
                        {
                            if (zone.InPoly2D(p.X, p.Y, p.Z))
                            {
                                cont = true;
                                break;
                            }
                        }

                        if (!cont)
                            continue;

                        //Check its not deselected

                        foreach (var zone in deselectPolysToCheck)
                        {
                            if (zone.InPoly2D(p.X, p.Y, p.Z))
                            {
                                cont = false;
                                break;
                            }
                        }

                        if (!cont)
                            continue;

                        //Point is selected and not deselected so add it to the section
                        double newX = 0, newY = 0;
                        Trig.DistanceAndOffsetFromLine(section.SectionCenter.X, section.SectionCenter.Y, section.Bearing, p.X, p.Y, ref newX, ref newY);
                        section.OriginalPoints.Add(new SectionPoint(newX, p.Z, newY, p.R, p.G, p.B));
                        section.SectionBounds.Update(newX, p.Z, newY);
                    }
                }

                points.Clear();

                if (Children != null)
                {
                    foreach (var child in Children)
                        if (child != null)
                            child.GetSectionFromSelectionWithoutClosest(ref section, selectionZones, deselectionZones, ref worker, e);
                }

                return true;
            }
            else
                return false;
        }

        public bool GetNextExtrudePoint(Ray ray, Polygon poly, List<Point3D> existingPoints, ref double x, ref double y, ref double z, ref byte r, ref byte g, ref byte b, ref ushort intensity, ref double distance)
        {
            if (Bounds.IntersectsWithRay(ray))
            {
                var points = LoadIntoStore();
                foreach (var p in points)
                {
                    if (poly.SquareBounds.IsInside(p.X, p.Y, p.Z))
                    {
                        if (poly.InPoly2D(p.X, p.Y, p.Z))
                        {
                            var newDist = Vector.ClosestDistanceOnALine(ray.Origin.X, ray.Origin.Y, ray.Origin.Z, ray.End.X, ray.End.Y, ray.End.Z, p.X, p.Y, p.Z);
                            if (newDist < distance)
                            {
                                bool add = true;
                                foreach (var point in existingPoints)
                                {
                                    if (p.X == point.X && p.Y == point.Y)
                                    {
                                        add = false;
                                        break;
                                    }
                                }

                                if (add)
                                {
                                    x = p.X;
                                    y = p.Y;
                                    z = p.Z;
                                    r = p.R;
                                    g = p.G;
                                    b = p.B;
                                    intensity = p.Intensity;
                                    distance = newDist;
                                }
                            }
                        }
                    }
                }
                points.Clear();
                if (Children != null)
                {
                    foreach (var child in Children)
                    {
                        if (child != null)
                            child.GetNextExtrudePoint(ray, poly, existingPoints, ref x, ref y, ref z, ref r, ref g, ref b, ref intensity, ref distance);
                    }
                }
                if (x == 0 && y == 0 && z == 0)
                    return false;
                else
                    return true;
            }
            else
                return false;
        }

        public bool GetSectionWithClosest(Ray ray, ref CrossSection section, ref double distance, ref BackgroundWorker worker, DoWorkEventArgs e)
        {
            //First get all points within the section bounds
            if (Bounds.Intersects2D(section.Poly.SquareBounds))
            {
                if (worker.CancellationPending == true)
                {
                    e.Cancel = true;
                    return false;
                }

                //Get all points within the section
                var points = LoadIntoStore();
                if (Bounds.IntersectsWithRay(ray))
                {
                    foreach (var p in points)
                    {
                        if (section.Poly.InPoly2D(p.X, p.Y, p.Z))
                        {
                            var newDist = Vector.ClosestDistanceOnALine(ray.Origin.X, ray.Origin.Y, ray.Origin.Z, ray.End.X, ray.End.Y, ray.End.Z, p.X, p.Y, p.Z);
                            if (newDist < distance)
                            {
                                section.ClosestPoint = new CloudPoint(p.X, p.Y, p.Z, p.R, p.G, p.B);
                                distance = newDist;
                            }

                            double newX = 0, newY = 0;
                            Trig.DistanceAndOffsetFromLine(section.SectionCenter.X, section.SectionCenter.Y, section.Bearing, p.X, p.Y, ref newX, ref newY);
                            section.SectionBounds.Update(newX, p.Z, newY);
                            section.OriginalPoints.Add(new SectionPoint(newX, p.Z, newY, p.R, p.G, p.B));
                        }
                    }
                }
                else
                {
                    foreach (var p in points)
                    {
                        if (section.Poly.InPoly2D(p.X, p.Y, p.Z))
                        {
                            double newX = 0, newY = 0;
                            Trig.DistanceAndOffsetFromLine(section.SectionCenter.X, section.SectionCenter.Y, section.Bearing, p.X, p.Y, ref newX, ref newY);
                            section.OriginalPoints.Add(new SectionPoint(newX, p.Z, newY, p.R, p.G, p.B));
                            section.SectionBounds.Update(newX, p.Z, newY);
                        }
                    }
                }
                points.Clear();

                if (Children != null)
                {
                    foreach (var child in Children)
                    {
                        if (child != null)
                            child.GetSectionWithClosest(ray, ref section, ref distance, ref worker, e);
                    }
                }

                return true;
            }
            else
                return false;
        }

        public bool GetSectionWithoutClosest(ref CrossSection section, ref BackgroundWorker worker, DoWorkEventArgs e, Alignment alignment = null)
        {
            //First get all points within the section bounds
            if (Bounds.Intersects2D(section.Poly.SquareBounds))
            {
                if (worker.CancellationPending == true)
                {
                    e.Cancel = true;
                    return false;
                }

                //Get all points within the section
                var points = LoadIntoStore();
                foreach (var p in points)
                {
                    if (section.Poly.InPoly2D(p.X, p.Y, p.Z))
                    {
                        if(alignment != null && section.Chainage != 1E20)
                        {
                            //Should only get an alignment when chainage is valid
                            var outVec = new AlignmentVector5();
                            if(alignment.Horizontal.GetChainageAndOffset(p, ref outVec))
                            {
                                section.OriginalPoints.Add(new SectionPoint(outVec.offset, p.Z, outVec.chainage - section.Chainage, p.R, p.G, p.B));
                                section.SectionBounds.Update(outVec.offset, p.Z, outVec.chainage - section.Chainage);
                            }
                        }
                        else
                        {
                            double newX = 0, newY = 0;
                            Trig.DistanceAndOffsetFromLine(section.SectionCenter.X, section.SectionCenter.Y, section.Bearing, p.X, p.Y, ref newX, ref newY);
                            section.OriginalPoints.Add(new SectionPoint(newX, p.Z, newY, p.R, p.G, p.B));
                            section.SectionBounds.Update(newX, p.Z, newY);
                        }
                    }
                }

                points.Clear();

                if (Children != null)
                {
                    foreach (var child in Children)
                    {
                        if (child != null)
                            child.GetSectionWithoutClosest(ref section, ref worker, e);
                    }
                }
                return true;
            }
            else
                return false;
        }

        public List<CloudPoint> GetSelectedPoints(List<Polygon> selectionZones, List<Polygon> deselectionZones)
        {
            //Check if bounds intersects with any selection zones
            var selectedPoints = new List<CloudPoint>();
            var polysToCheck = new List<Polygon>();
            var deselectPolysToCheck = new List<Polygon>();

            try
            {
                //Find the selection zones that intersect with this node
                foreach (var poly in selectionZones)
                {
                    //If the zone is filtered we don't need points in it anyway
                    if (poly.Filtered)
                        continue;

                    if (Bounds.Intersects2D(poly.SquareBounds))
                        polysToCheck.Add(poly);
                }

                //If there are no selected zones that intersect then there are no selected points so return
                if (polysToCheck.Count <= 0)
                    return selectedPoints;

                //Find the deselection zones that intersect with this node
                foreach(var poly in deselectionZones)
                {
                    if (Bounds.Intersects2D(poly.SquareBounds))
                        deselectPolysToCheck.Add(poly);
                }

                //Collect the points for the node
                var points = LoadIntoStore();

                foreach(var point in points)
                {
                    bool pointSelected = false;

                    //Check if the point falls within a selection zone
                    foreach(var poly in polysToCheck)
                    {
                        if(poly.InPoly2D(point.X, point.Y, point.Z))
                        {
                            pointSelected = true;

                            //Check if the point falls within a deselection zone
                            foreach(var deselectPoly in deselectPolysToCheck)
                            {
                                if(deselectPoly.InPoly2D(point.X, point.Y, point.Z))
                                {
                                    pointSelected = false;
                                    break;
                                }
                            }

                            break;
                        }
                    }

                    if (!pointSelected)
                        continue;

                    //Now that we know the point is selected and is not deselected we can add it to the list
                    selectedPoints.Add(point);
                }

                return selectedPoints;
            }
            catch
            {
                return selectedPoints;
            }
        }

        #endregion
    }

    public class ConversionProgress
    {
        #region Properties

        public ConversionStage Stage { get; set; }
        public int PrimaryPercentage { get; set; }
        public int SecondaryPercentrage { get; set; }
        public string SecondaryMessage { get; set; }
        public ulong TotalPoints { get; set; }
        public ulong CurrentPoints { get; set; }
        public ulong TotalFilePoints { get; set; }
        public ulong currentFilePoints { get; set; }

        #endregion

        #region Setup

        public ConversionProgress()
        {
            Stage = ConversionStage.PreparingFiles;
            PrimaryPercentage = 0;
            SecondaryPercentrage = 0;
            SecondaryMessage = "";
            TotalPoints = 0;
            CurrentPoints = 0;
            TotalFilePoints = 0;
            currentFilePoints = 0;
        }

        #endregion
    }

    public enum ConversionStage
    {
        PreparingFiles,
        GetBounds,
        BuildingOctree,
        FlushingOctree,
        FinalizingOctree
    }
}
